import json
import logging

from markupsafe import escape as escape_html
from sqlalchemy import (
    and_,
    false,
    or_,
    true,
)

import tool_shed.grids.util as grids_util
import tool_shed.repository_types.util as rt_util
import tool_shed.util.shed_util_common as suc
from galaxy.web.legacy_framework import grids
from tool_shed.util import (
    hg_util,
    metadata_util,
)
from tool_shed.webapp import model
from tool_shed.webapp.model.db import get_repository_by_name_and_owner

log = logging.getLogger(__name__)


class CategoryGrid(grids.Grid):
    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, category):
            return category.name

    class DescriptionColumn(grids.TextColumn):
        def get_value(self, trans, grid, category):
            return category.description

    class RepositoriesColumn(grids.TextColumn):
        def get_value(self, trans, grid, category):
            category_name = str(category.name)
            filter = trans.app.repository_grid_filter_manager.get_filter(trans)
            if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
                return (
                    trans.app.repository_registry.certified_level_one_viewable_repositories_and_suites_by_category.get(
                        category_name, 0
                    )
                )
            elif filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
                return trans.app.repository_registry.certified_level_one_viewable_suites_by_category.get(
                    category_name, 0
                )
            elif filter == trans.app.repository_grid_filter_manager.filters.SUITES:
                return trans.app.repository_registry.viewable_suites_by_category.get(category_name, 0)
            else:
                # The value filter is None.
                return trans.app.repository_registry.viewable_repositories_and_suites_by_category.get(category_name, 0)

    title = "Categories"
    model_class = model.Category
    template = "/webapps/tool_shed/category/grid.mako"
    default_sort_key = "name"
    columns = [
        NameColumn(
            "Name",
            key="Category.name",
            link=(lambda item: dict(operation="repositories_by_category", id=item.id)),
            attach_popup=False,
        ),
        DescriptionColumn("Description", key="Category.description", attach_popup=False),
        RepositoriesColumn("Repositories", model_class=model.Repository, attach_popup=False),
    ]
    # Override these
    num_rows_per_page = 50


class RepositoryGrid(grids.Grid):
    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            return escape_html(repository.name)

    class TypeColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            type_class = repository.get_type_class(trans.app)
            return escape_html(type_class.label)

    class HeadsColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            """Display the current repository heads."""
            repo = repository.hg_repo
            heads = hg_util.get_repository_heads(repo)
            multiple_heads = len(heads) > 1
            if multiple_heads:
                heads_str = '<font color="red">'
            else:
                heads_str = ""
            for ctx in heads:
                heads_str += f"{hg_util.get_revision_label_from_ctx(ctx, include_date=True)}<br/>"
            heads_str.rstrip("<br/>")
            if multiple_heads:
                heads_str += "</font>"
            return heads_str

    class MetadataRevisionColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            """Display a SelectField whose options are the changeset_revision strings of all metadata revisions of this repository."""
            # A repository's metadata revisions may not all be installable, as some may contain only invalid tools.
            select_field = grids_util.build_changeset_revision_select_field(trans, repository, downloadable=False)
            if len(select_field.options) > 1:
                tmpl = f"<select name='{select_field.name}'>"
                for o in select_field.options:
                    tmpl += f"<option value='{o[1]}'>{o[0]}</option>"
                tmpl += "</select>"
                return tmpl
            elif len(select_field.options) == 1:
                option_items = select_field.options[0][0]
                rev_label, rev_date = option_items.split(" ")
                rev_date = f'<i><font color="#666666">{rev_date}</font></i>'
                return f"{rev_label} {rev_date}"
            return ""

    class LatestInstallableRevisionColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            """Display the latest installable revision label (may not be the repository tip)."""
            select_field = grids_util.build_changeset_revision_select_field(trans, repository, downloadable=False)
            if select_field.options:
                return select_field.options[0][0]
            return ""

    class TipRevisionColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            """Display the repository tip revision label."""
            return escape_html(repository.revision())

    class DescriptionColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            return escape_html(repository.description)

    class CategoryColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            rval = "<ul>"
            if repository.categories:
                for rca in repository.categories:
                    rval += f'<li><a href="browse_repositories?operation=repositories_by_category&id={trans.security.encode_id(rca.category.id)}">{rca.category.name}</a></li>'
            else:
                rval += "<li>not set</li>"
            rval += "</ul>"
            return rval

    class RepositoryCategoryColumn(grids.GridColumn):
        def filter(self, trans, user, query, column_filter):
            """Modify query to filter by category."""
            if column_filter == "All":
                return query
            return query.filter(model.Category.name == column_filter)

    class UserColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            if repository.user:
                return escape_html(repository.user.username)
            return "no user"

    class EmailColumn(grids.TextColumn):
        def filter(self, trans, user, query, column_filter):
            if column_filter == "All":
                return query
            return query.filter(
                and_(
                    model.Repository.table.c.user_id == model.User.table.c.id, model.User.table.c.email == column_filter
                )
            )

    class EmailAlertsColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            if trans.user and repository.email_alerts and trans.user.email in json.loads(repository.email_alerts):
                return "yes"
            return ""

    class DeprecatedColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            if repository.deprecated:
                return "yes"
            return ""

    title = "Repositories"
    model_class = model.Repository
    default_sort_key = "name"
    use_hide_message = False
    columns = [
        NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        DescriptionColumn("Synopsis", key="description", attach_popup=False),
        TypeColumn("Type"),
        MetadataRevisionColumn("Metadata<br/>Revisions"),
        UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
        # Columns that are valid for filtering but are not visible.
        EmailColumn("Email", model_class=model.User, key="email", visible=False),
        RepositoryCategoryColumn("Category", model_class=model.Category, key="Category.name", visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, description",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    default_filter = dict(deleted="False")
    num_rows_per_page = 50
    use_paging = True
    allow_fetching_all_results = False

    def build_initial_query(self, trans, **kwd):
        filter = trans.app.repository_grid_filter_manager.get_filter(trans)
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
            return (
                trans.sa_session.query(model.Repository)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
            return (
                trans.sa_session.query(model.Repository)
                .filter(model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        else:
            # The filter is None.
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false())
                )
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )


class DockerImageGrid(RepositoryGrid):
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.DescriptionColumn("Synopsis", key="description", attach_popup=False),
        RepositoryGrid.UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
        RepositoryGrid.EmailAlertsColumn("Alert", attach_popup=False),
    ]
    operations = [grids.GridOperation("Include in Docker image", allow_multiple=True)]
    show_item_checkboxes = True


class EmailAlertsRepositoryGrid(RepositoryGrid):
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.DescriptionColumn("Synopsis", key="description", attach_popup=False),
        RepositoryGrid.UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
        RepositoryGrid.EmailAlertsColumn("Alert", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
    ]
    operations = [grids.GridOperation("Receive email alerts", allow_multiple=True)]
    global_actions = [
        grids.GridAction("User preferences", dict(controller="user", action="index", cntrller="repository"))
    ]


class MatchedRepositoryGrid(grids.Grid):
    # This grid filters out repositories that have been marked as deleted or deprecated.

    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            return escape_html(repository_metadata.repository.name)

    class DescriptionColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            return escape_html(repository_metadata.repository.description)

    class RevisionColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            return repository_metadata.changeset_revision

    class UserColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.repository.user:
                return escape_html(repository_metadata.repository.user.username)
            return "no user"

    # Grid definition
    title = "Matching repositories"
    model_class = model.RepositoryMetadata
    default_sort_key = "Repository.name"
    use_hide_message = False
    columns = [
        NameColumn(
            "Repository name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=True,
        ),
        DescriptionColumn("Synopsis", attach_popup=False),
        RevisionColumn("Revision"),
        UserColumn("Owner", model_class=model.User, attach_popup=False),
    ]
    operations = [grids.GridOperation("Install to Galaxy", allow_multiple=True)]
    num_rows_per_page = 50
    use_paging = False

    def build_initial_query(self, trans, **kwd):
        clause_list = []
        if match_tuples := kwd.get("match_tuples", []):
            for match_tuple in match_tuples:
                repository_id, changeset_revision = match_tuple
                clause_list.append(
                    and_(
                        model.RepositoryMetadata.repository_id == int(repository_id),
                        model.RepositoryMetadata.changeset_revision == changeset_revision,
                    )
                )
            return (
                trans.sa_session.query(model.RepositoryMetadata)
                .join(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false())
                )
                .join(model.User.table)
                .filter(or_(*clause_list))
                .order_by(model.Repository.name)
            )
        # Return an empty query
        return trans.sa_session.query(model.RepositoryMetadata).filter(model.RepositoryMetadata.id < 0)


class InstallMatchedRepositoryGrid(MatchedRepositoryGrid):
    columns = list(MatchedRepositoryGrid.columns)
    # Override the NameColumn
    columns[0] = MatchedRepositoryGrid.NameColumn(
        "Name", link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)), attach_popup=False
    )


class MyWritableRepositoriesGrid(RepositoryGrid):
    # This grid filters out repositories that have been marked as either deprecated or deleted.
    title = "Repositories I can change"
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.TypeColumn("Type"),
        RepositoryGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoryGrid.UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        # TODO: improve performance by adding a db table associating users with repositories for which they have write access.
        username = trans.user.username
        clause_list = []
        for repository in trans.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
        ):
            allow_push = repository.allow_push()
            if allow_push:
                allow_push_usernames = allow_push.split(",")
                if username in allow_push_usernames:
                    clause_list.append(model.Repository.table.c.id == repository.id)
        if clause_list:
            return trans.sa_session.query(model.Repository).filter(or_(*clause_list)).join(model.User.table)
        # Return an empty query.
        return trans.sa_session.query(model.Repository).filter(model.Repository.table.c.id < 0)


class RepositoriesByUserGrid(RepositoryGrid):
    title = "Repositories by user"
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.DescriptionColumn("Synopsis", key="description", attach_popup=False),
        RepositoryGrid.TypeColumn("Type"),
        RepositoryGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoryGrid.CategoryColumn("Category", model_class=model.Category, key="Category.name", attach_popup=False),
    ]
    default_filter = dict(deleted="False")

    def build_initial_query(self, trans, **kwd):
        decoded_user_id = trans.security.decode_id(kwd["user_id"])
        filter = trans.app.repository_grid_filter_manager.get_filter(trans)
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
            return (
                trans.sa_session.query(model.Repository)
                .filter(model.Repository.table.c.user_id == decoded_user_id)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(
                        model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION,
                        model.Repository.table.c.user_id == decoded_user_id,
                    )
                )
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        else:
            # The value of filter is None.
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(
                        model.Repository.table.c.deleted == false(),
                        model.Repository.table.c.deprecated == false(),
                        model.Repository.table.c.user_id == decoded_user_id,
                    )
                )
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )


class RepositoriesInCategoryGrid(RepositoryGrid):
    title = "Category"

    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(controller="repository", operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.DescriptionColumn("Synopsis", key="description", attach_popup=False),
        RepositoryGrid.TypeColumn("Type"),
        RepositoryGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoryGrid.UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(controller="repository", operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
        # Columns that are valid for filtering but are not visible.
        RepositoryGrid.EmailColumn("Email", model_class=model.User, key="email", visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, description",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        category_id = kwd.get("id", None)
        filter = trans.app.repository_grid_filter_manager.get_filter(trans)
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
            if category_id:
                category = suc.get_category(trans.app, category_id)
                if category:
                    return (
                        trans.sa_session.query(model.Repository)
                        .join(model.RepositoryMetadata.table)
                        .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                        .join(model.User.table)
                        .outerjoin(model.RepositoryCategoryAssociation)
                        .outerjoin(model.Category.table)
                        .filter(model.Category.table.c.name == category.name)
                    )
            return (
                trans.sa_session.query(model.Repository)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
            if category_id:
                category = suc.get_category(trans.app, category_id)
                if category:
                    return (
                        trans.sa_session.query(model.Repository)
                        .filter(model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION)
                        .join(model.RepositoryMetadata.table)
                        .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                        .join(model.User.table)
                        .outerjoin(model.RepositoryCategoryAssociation)
                        .outerjoin(model.Category.table)
                        .filter(model.Category.table.c.name == category.name)
                    )
            return (
                trans.sa_session.query(model.Repository)
                .filter(model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )
        else:
            # The value of filter is None.
            if category_id:
                category = suc.get_category(trans.app, category_id)
                if category:
                    return (
                        trans.sa_session.query(model.Repository)
                        .filter(
                            and_(
                                model.Repository.table.c.deleted == false(),
                                model.Repository.table.c.deprecated == false(),
                            )
                        )
                        .join(model.User.table)
                        .outerjoin(model.RepositoryCategoryAssociation)
                        .outerjoin(model.Category.table)
                        .filter(model.Category.table.c.name == category.name)
                    )
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false())
                )
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
            )


class RepositoriesIOwnGrid(RepositoryGrid):
    title = "Repositories I own"
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.TypeColumn("Type"),
        RepositoryGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoryGrid.DeprecatedColumn("Deprecated"),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.Repository)
            .filter(
                and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.user_id == trans.user.id)
            )
            .join(model.User.table)
            .outerjoin(model.RepositoryCategoryAssociation)
            .outerjoin(model.Category.table)
        )


class RepositoriesICanAdministerGrid(RepositoryGrid):
    title = "Repositories I can administer"
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.UserColumn("Owner"),
        RepositoryGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoryGrid.DeprecatedColumn("Deprecated"),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        """
        Retrieve all repositories for which the current user has been granted administrative privileges.
        """
        current_user = trans.user
        # Build up an or-based clause list containing role table records.
        clause_list = []
        # Include each of the user's roles.
        for ura in current_user.roles:
            clause_list.append(model.Role.table.c.id == ura.role_id)
        # Include each role associated with each group of which the user is a member.
        for uga in current_user.groups:
            group = uga.group
            for gra in group.roles:
                clause_list.append(model.Role.table.c.id == gra.role_id)
        # Filter out repositories for which the user does not have the administrative role either directly
        # via a role association or indirectly via a group -> role association.
        return (
            trans.sa_session.query(model.Repository)
            .filter(model.Repository.table.c.deleted == false())
            .outerjoin(model.RepositoryRoleAssociation)
            .outerjoin(model.Role.table)
            .filter(or_(*clause_list))
            .join(model.User.table)
            .outerjoin(model.RepositoryCategoryAssociation)
            .outerjoin(model.Category.table)
        )


class RepositoriesMissingToolTestComponentsGrid(RepositoryGrid):
    # This grid displays only the latest installable revision of each repository.
    title = "Repositories with missing tool test components"
    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.LatestInstallableRevisionColumn("Latest Installable Revision"),
        RepositoryGrid.UserColumn(
            "Owner",
            key="User.username",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
        ),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        # Filter by latest installable revisions that contain tools with missing tool test components.
        revision_clause_list = []
        for repository in trans.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
        ):
            changeset_revision = (
                grids_util.filter_by_latest_downloadable_changeset_revision_that_has_missing_tool_test_components(
                    trans, repository
                )
            )
            if changeset_revision:
                revision_clause_list.append(model.RepositoryMetadata.table.c.changeset_revision == changeset_revision)
        if revision_clause_list:
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
                )
                .join(model.RepositoryMetadata)
                .filter(or_(*revision_clause_list))
                .join(model.User.table)
            )
        # Return an empty query.
        return trans.sa_session.query(model.Repository).filter(model.Repository.table.c.id < 0)


class MyWritableRepositoriesMissingToolTestComponentsGrid(RepositoriesMissingToolTestComponentsGrid):
    # This grid displays only the latest installable revision of each repository.
    title = "Repositories I can change with missing tool test components"
    columns = list(RepositoriesMissingToolTestComponentsGrid.columns)

    def build_initial_query(self, trans, **kwd):
        # First get all repositories that the current user is authorized to update.
        username = trans.user.username
        user_clause_list = []
        for repository in trans.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
        ):
            allow_push = repository.allow_push()
            if allow_push:
                allow_push_usernames = allow_push.split(",")
                if username in allow_push_usernames:
                    user_clause_list.append(model.Repository.table.c.id == repository.id)
        if user_clause_list:
            # We have the list of repositories that the current user is authorized to update, so filter
            # further by latest installable revisions that contain tools with missing tool test components.
            revision_clause_list = []
            for repository in (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
                )
                .filter(or_(*user_clause_list))
            ):
                changeset_revision = (
                    grids_util.filter_by_latest_downloadable_changeset_revision_that_has_missing_tool_test_components(
                        trans, repository
                    )
                )
                if changeset_revision:
                    revision_clause_list.append(
                        model.RepositoryMetadata.table.c.changeset_revision == changeset_revision
                    )
            if revision_clause_list:
                return (
                    trans.sa_session.query(model.Repository)
                    .filter(
                        and_(
                            model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false()
                        )
                    )
                    .join(model.User.table)
                    .filter(or_(*user_clause_list))
                    .join(model.RepositoryMetadata)
                    .filter(or_(*revision_clause_list))
                )
        # Return an empty query.
        return trans.sa_session.query(model.Repository).filter(model.Repository.table.c.id < 0)


class DeprecatedRepositoriesIOwnGrid(RepositoriesIOwnGrid):
    title = "Deprecated repositories I own"
    columns = [
        RepositoriesIOwnGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.TypeColumn("Type"),
        RepositoriesIOwnGrid.MetadataRevisionColumn("Metadata<br/>Revisions"),
        RepositoriesIOwnGrid.CategoryColumn(
            "Category", model_class=model.Category, key="Category.name", attach_popup=False
        ),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.Repository)
            .filter(
                and_(
                    model.Repository.table.c.deleted == false(),
                    model.Repository.table.c.user_id == trans.user.id,
                    model.Repository.table.c.deprecated == true(),
                )
            )
            .join(model.User.table)
            .outerjoin(model.RepositoryCategoryAssociation)
            .outerjoin(model.Category.table)
        )


class RepositoriesWithInvalidToolsGrid(RepositoryGrid):
    # This grid displays only the latest installable revision of each repository.

    class InvalidToolConfigColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            # At the time this grid is displayed we know that the received repository will have invalid tools in its latest changeset revision
            # that has associated metadata.
            val = ""
            repository_metadata = grids_util.get_latest_repository_metadata_if_it_includes_invalid_tools(
                trans, repository
            )
            metadata = repository_metadata.metadata
            if invalid_tools := metadata.get("invalid_tools", []):
                for invalid_tool_config in invalid_tools:
                    href_str = f'<a href="load_invalid_tool?repository_id={trans.security.encode_id(repository.id)}&tool_config={invalid_tool_config}&changeset_revision={repository_metadata.changeset_revision}">{invalid_tool_config}</a>'
                    val += href_str
                    val += "<br/>"
                val = val.rstrip("<br/>")
            return val

    title = "Repositories with invalid tools"
    columns = [
        InvalidToolConfigColumn("Tool config"),
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryGrid.LatestInstallableRevisionColumn("Latest Metadata Revision"),
        RepositoryGrid.UserColumn(
            "Owner",
            key="User.username",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
        ),
    ]

    def build_initial_query(self, trans, **kwd):
        # Filter by latest metadata revisions that contain invalid tools.
        revision_clause_list = []
        for repository in trans.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
        ):
            changeset_revision = grids_util.filter_by_latest_metadata_changeset_revision_that_has_invalid_tools(
                trans, repository
            )
            if changeset_revision:
                revision_clause_list.append(model.RepositoryMetadata.table.c.changeset_revision == changeset_revision)
        if revision_clause_list:
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
                )
                .join(model.RepositoryMetadata)
                .filter(or_(*revision_clause_list))
                .join(model.User.table)
            )
        # Return an empty query.
        return trans.sa_session.query(model.Repository).filter(model.Repository.table.c.id < 0)


class MyWritableRepositoriesWithInvalidToolsGrid(RepositoriesWithInvalidToolsGrid):
    # This grid displays only the latest installable revision of each repository.
    title = "Repositories I can change with invalid tools"
    columns = list(RepositoriesWithInvalidToolsGrid.columns)

    def build_initial_query(self, trans, **kwd):
        # First get all repositories that the current user is authorized to update.
        username = trans.user.username
        user_clause_list = []
        for repository in trans.sa_session.query(model.Repository).filter(
            and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
        ):
            allow_push = repository.allow_push()
            if allow_push:
                allow_push_usernames = allow_push.split(",")
                if username in allow_push_usernames:
                    user_clause_list.append(model.Repository.table.c.id == repository.id)
        if user_clause_list:
            # We have the list of repositories that the current user is authorized to update, so filter
            # further by latest metadata revisions that contain invalid tools.
            revision_clause_list = []
            for repository in (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false())
                )
                .filter(or_(*user_clause_list))
            ):
                changeset_revision = grids_util.filter_by_latest_metadata_changeset_revision_that_has_invalid_tools(
                    trans, repository
                )
                if changeset_revision:
                    revision_clause_list.append(
                        model.RepositoryMetadata.table.c.changeset_revision == changeset_revision
                    )
            if revision_clause_list:
                return (
                    trans.sa_session.query(model.Repository)
                    .filter(
                        and_(
                            model.Repository.table.c.deprecated == false(), model.Repository.table.c.deleted == false()
                        )
                    )
                    .join(model.User.table)
                    .filter(or_(*user_clause_list))
                    .join(model.RepositoryMetadata)
                    .filter(or_(*revision_clause_list))
                )
        # Return an empty query.
        return trans.sa_session.query(model.Repository).filter(model.Repository.table.c.id < 0)


class RepositoryMetadataGrid(grids.Grid):
    class RepositoryNameColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            repository = repository_metadata.repository
            return escape_html(repository.name)

    class RepositoryTypeColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            repository = repository_metadata.repository
            type_class = repository.get_type_class(trans.app)
            return escape_html(type_class.label)

    class RepositoryOwnerColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            repository = repository_metadata.repository
            return escape_html(repository.user.username)

    class ChangesetRevisionColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            repository = repository_metadata.repository
            changeset_revision = repository_metadata.changeset_revision
            changeset_revision_label = hg_util.get_revision_label(
                trans.app, repository, changeset_revision, include_date=True
            )
            return changeset_revision_label

    class MaliciousColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.malicious:
                return "yes"
            return ""

    class DownloadableColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.downloadable:
                return "yes"
            return ""

    class HasRepositoryDependenciesColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.has_repository_dependencies:
                return "yes"
            return ""

    class IncludesDatatypesColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.includes_datatypes:
                return "yes"
            return ""

    class IncludesToolsColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.includes_tools:
                return "yes"
            return ""

    class IncludesToolDependenciesColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.includes_tool_dependencies:
                return "yes"
            return ""

    class IncludesWorkflowsColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.includes_workflows:
                return "yes"
            return ""

    title = "Repository metadata"
    model_class = model.RepositoryMetadata
    default_sort_key = "Repository.name"
    columns = [
        RepositoryNameColumn(
            "Repository name",
            key="Repository.name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
        ),
        RepositoryNameColumn("Type"),
        RepositoryOwnerColumn("Owner", model_class=model.User, attach_popup=False, key="User.username"),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, description",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    default_filter = dict(malicious="False")
    num_rows_per_page = 50
    use_paging = True
    allow_fetching_all_results = False

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.RepositoryMetadata)
            .join(model.Repository)
            .filter(and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false()))
            .join(model.User.table)
        )


class RepositoryDependenciesGrid(RepositoryMetadataGrid):
    class RequiredRepositoryColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            rd_str = []
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    rd_dict = metadata.get("repository_dependencies", {})
                    if rd_dict:
                        rd_tups = rd_dict["repository_dependencies"]
                        # "repository_dependencies": [["http://localhost:9009", "bwa059", "test", "a07baa797d53"]]
                        # Sort rd_tups by by required repository name.
                        sorted_rd_tups = sorted(rd_tups, key=lambda rd_tup: rd_tup[1])
                        for rd_tup in sorted_rd_tups:
                            name, owner, changeset_revision = rd_tup[1:4]
                            rd_line = ""
                            required_repository = get_repository_by_name_and_owner(trans.sa_session, name, owner)
                            if required_repository and not required_repository.deleted:
                                required_repository_id = trans.security.encode_id(required_repository.id)
                                required_repository_metadata = (
                                    metadata_util.get_repository_metadata_by_repository_id_changeset_revision(
                                        trans.app, required_repository_id, changeset_revision
                                    )
                                )
                                if not required_repository_metadata:
                                    updated_changeset_revision = metadata_util.get_next_downloadable_changeset_revision(
                                        trans.app, required_repository, changeset_revision
                                    )
                                    required_repository_metadata = (
                                        metadata_util.get_repository_metadata_by_repository_id_changeset_revision(
                                            trans.app, required_repository_id, updated_changeset_revision
                                        )
                                    )
                                required_repository_metadata_id = trans.security.encode_id(
                                    required_repository_metadata.id
                                )
                                rd_line += f'<a href="browse_repository_dependencies?operation=view_or_manage_repository&id={required_repository_metadata_id}">'
                            rd_line += f"Repository <b>{escape_html(name)}</b> revision <b>{escape_html(owner)}</b> owned by <b>{escape_html(changeset_revision)}</b>"
                            if required_repository:
                                rd_line += "</a>"
                            rd_str.append(rd_line)
            return "<br />".join(rd_str)

    title = "Valid repository dependency definitions in this tool shed"
    default_sort_key = "Repository.name"
    columns = [
        RequiredRepositoryColumn("Repository dependency", attach_popup=False),
        RepositoryMetadataGrid.RepositoryNameColumn(
            "Repository name",
            model_class=model.Repository,
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
            key="Repository.name",
        ),
        RepositoryMetadataGrid.RepositoryOwnerColumn(
            "Owner", model_class=model.User, attach_popup=False, key="User.username"
        ),
        RepositoryMetadataGrid.ChangesetRevisionColumn("Revision", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, owner",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.RepositoryMetadata)
            .join(model.Repository)
            .filter(
                and_(
                    model.RepositoryMetadata.table.c.has_repository_dependencies == true(),
                    model.Repository.table.c.deleted == false(),
                    model.Repository.table.c.deprecated == false(),
                )
            )
            .join(model.User.table)
        )


class DatatypesGrid(RepositoryMetadataGrid):
    class DatatypesColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            datatype_list = []
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    datatype_dicts = metadata.get("datatypes", [])
                    if datatype_dicts:
                        # Create tuples of the attributes we want so we can sort them by extension.
                        datatype_tups = []
                        for datatype_dict in datatype_dicts:
                            # Example: {"display_in_upload": "true", "dtype": "galaxy.datatypes.blast:BlastXml", "extension": "blastxml", "mimetype": "application/xml"}
                            extension = datatype_dict.get("extension", "")
                            dtype = datatype_dict.get("dtype", "")
                            # For now we'll just display extension and dtype.
                            if extension and dtype:
                                datatype_tups.append((extension, dtype))
                        sorted_datatype_tups = sorted(datatype_tups, key=lambda datatype_tup: datatype_tup[0])
                        for datatype_tup in sorted_datatype_tups:
                            extension, datatype = datatype_tup[:2]
                            datatype_str = f'<a href="browse_datatypes?operation=view_or_manage_repository&id={trans.security.encode_id(repository_metadata.id)}">'
                            datatype_str += f"<b>{escape_html(extension)}:</b> {escape_html(datatype)}"
                            datatype_str += "</a>"
                            datatype_list.append(datatype_str)
            return "<br />".join(datatype_list)

    title = "Custom datatypes in this tool shed"
    default_sort_key = "Repository.name"
    columns = [
        DatatypesColumn("Datatype extension and class", attach_popup=False),
        RepositoryMetadataGrid.RepositoryNameColumn(
            "Repository name",
            model_class=model.Repository,
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
            key="Repository.name",
        ),
        RepositoryMetadataGrid.RepositoryOwnerColumn(
            "Owner", model_class=model.User, attach_popup=False, key="User.username"
        ),
        RepositoryMetadataGrid.ChangesetRevisionColumn("Revision", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, owner",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.RepositoryMetadata)
            .join(model.Repository)
            .filter(
                and_(
                    model.RepositoryMetadata.table.c.includes_datatypes == true(),
                    model.Repository.table.c.deleted == false(),
                    model.Repository.table.c.deprecated == false(),
                )
            )
            .join(model.User.table)
        )


class ToolDependenciesGrid(RepositoryMetadataGrid):
    class ToolDependencyColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            td_str = ""
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    tds_dict = metadata.get("tool_dependencies", {})
                    if tds_dict:
                        # Example: {"bwa/0.5.9": {"name": "bwa", "type": "package", "version": "0.5.9"}}
                        sorted_keys = sorted(tds_dict.keys())
                        num_keys = len(sorted_keys)
                        # Handle environment settings first.
                        if "set_environment" in sorted_keys:
                            # Example: "set_environment": [{"name": "JAVA_JAR_FILE", "type": "set_environment"}]
                            env_dicts = tds_dict["set_environment"]
                            num_env_dicts = len(env_dicts)
                            if num_env_dicts > 0:
                                td_str += f'<a href="browse_datatypes?operation=view_or_manage_repository&id={trans.security.encode_id(repository_metadata.id)}">'
                                td_str += "<b>environment:</b> "
                                td_str += ", ".join(escape_html(env_dict["name"]) for env_dict in env_dicts)
                                td_str += "</a><br/>"
                        for index, key in enumerate(sorted_keys):
                            if key == "set_environment":
                                continue
                            td_dict = tds_dict[key]
                            # Example: {"name": "bwa", "type": "package", "version": "0.5.9"}
                            name = td_dict["name"]
                            version = td_dict["version"]
                            td_str += f'<a href="browse_datatypes?operation=view_or_manage_repository&id={trans.security.encode_id(repository_metadata.id)}">'
                            td_str += f"<b>{escape_html(name)}</b> version <b>{escape_html(version)}</b>"
                            td_str += "</a>"
                            if index < num_keys - 1:
                                td_str += "<br/>"
            return td_str

    title = "Tool dependency definitions in this tool shed"
    default_sort_key = "Repository.name"
    columns = [
        ToolDependencyColumn("Tool dependency", attach_popup=False),
        RepositoryMetadataGrid.RepositoryNameColumn(
            "Repository name",
            model_class=model.Repository,
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
            key="Repository.name",
        ),
        RepositoryMetadataGrid.RepositoryOwnerColumn(
            "Owner", model_class=model.User, attach_popup=False, key="User.username"
        ),
        RepositoryMetadataGrid.ChangesetRevisionColumn("Revision", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, owner",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.RepositoryMetadata)
            .join(model.Repository)
            .filter(
                and_(
                    model.RepositoryMetadata.table.c.includes_tool_dependencies == true(),
                    model.Repository.table.c.deleted == false(),
                    model.Repository.table.c.deprecated == false(),
                )
            )
            .join(model.User.table)
        )


class ToolsGrid(RepositoryMetadataGrid):
    class ToolsColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            tool_line = []
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    tool_dicts = metadata.get("tools", [])
                    if tool_dicts:
                        # Create tuples of the attributes we want so we can sort them by extension.
                        tool_tups = []
                        for tool_dict in tool_dicts:
                            tool_id = tool_dict.get("id", "")
                            version = tool_dict.get("version", "")
                            # For now we'll just display tool id and version.
                            if tool_id and version:
                                tool_tups.append((tool_id, version))
                        sorted_tool_tups = sorted(tool_tups, key=lambda tool_tup: tool_tup[0])
                        for tool_tup in sorted_tool_tups:
                            tool_id, version = tool_tup[:2]
                            tool_str = f'<a href="browse_datatypes?operation=view_or_manage_repository&id={trans.security.encode_id(repository_metadata.id)}">'
                            tool_str += f"<b>{escape_html(tool_id)}:</b> {escape_html(version)}"
                            tool_str += "</a>"
                            tool_line.append(tool_str)
            return "<br />".join(tool_line)

    title = "Valid tools in this tool shed"
    default_sort_key = "Repository.name"
    columns = [
        ToolsColumn("Tool id and version", attach_popup=False),
        RepositoryMetadataGrid.RepositoryNameColumn(
            "Repository name",
            model_class=model.Repository,
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=False,
            key="Repository.name",
        ),
        RepositoryMetadataGrid.RepositoryOwnerColumn(
            "Owner", model_class=model.User, attach_popup=False, key="User.username"
        ),
        RepositoryMetadataGrid.ChangesetRevisionColumn("Revision", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, owner",
            cols_to_filter=[columns[1], columns[2]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        return (
            trans.sa_session.query(model.RepositoryMetadata)
            .join(model.Repository)
            .filter(
                and_(
                    model.RepositoryMetadata.table.c.includes_tools == true(),
                    model.Repository.table.c.deleted == false(),
                    model.Repository.table.c.deprecated == false(),
                )
            )
            .join(model.User.table)
        )


class ValidCategoryGrid(CategoryGrid):
    class RepositoriesColumn(grids.TextColumn):
        def get_value(self, trans, grid, category):
            category_name = str(category.name)
            filter = trans.app.repository_grid_filter_manager.get_filter(trans)
            if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
                return (
                    trans.app.repository_registry.certified_level_one_viewable_repositories_and_suites_by_category.get(
                        category_name, 0
                    )
                )
            elif filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
                return trans.app.repository_registry.certified_level_one_viewable_suites_by_category.get(
                    category_name, 0
                )
            elif filter == trans.app.repository_grid_filter_manager.filters.SUITES:
                return trans.app.repository_registry.viewable_valid_suites_by_category.get(category_name, 0)
            else:
                # The value filter is None.
                return trans.app.repository_registry.viewable_valid_repositories_and_suites_by_category.get(
                    category_name, 0
                )

    title = "Categories of Valid Repositories"
    model_class = model.Category
    template = "/webapps/tool_shed/category/valid_grid.mako"
    default_sort_key = "name"
    columns = [
        CategoryGrid.NameColumn(
            "Name",
            key="Category.name",
            link=(lambda item: dict(operation="valid_repositories_by_category", id=item.id)),
            attach_popup=False,
        ),
        CategoryGrid.DescriptionColumn("Description", key="Category.description", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        RepositoriesColumn("Valid repositories", model_class=model.Repository, attach_popup=False),
    ]
    # Override these
    num_rows_per_page = 50


class ValidRepositoryGrid(RepositoryGrid):
    # This grid filters out repositories that have been marked as either deleted or deprecated.

    class CategoryColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository):
            rval = "<ul>"
            if repository.categories:
                for rca in repository.categories:
                    rval += f'<li><a href="browse_repositories?operation=valid_repositories_by_category&id={trans.security.encode_id(rca.category.id)}">{rca.category.name}</a></li>'
            else:
                rval += "<li>not set</li>"
            rval += "</ul>"
            return rval

    class RepositoryCategoryColumn(grids.GridColumn):
        def filter(self, trans, user, query, column_filter):
            """Modify query to filter by category."""
            if column_filter == "All":
                return query
            return query.filter(model.Category.name == column_filter)

    class InstallableRevisionColumn(grids.GridColumn):
        def __init__(self, col_name):
            grids.GridColumn.__init__(self, col_name)

        def get_value(self, trans, grid, repository):
            """Display a SelectField whose options are the changeset_revision strings of all download-able revisions of this repository."""
            select_field = grids_util.build_changeset_revision_select_field(trans, repository, downloadable=True)
            if len(select_field.options) > 1:
                tmpl = f"<select name='{select_field.name}'>"
                for o in select_field.options:
                    tmpl += f"<option value='{o[1]}'>{o[0]}</option>"
                tmpl += "</select>"
                return tmpl
            elif len(select_field.options) == 1:
                return select_field.options[0][0]
            return ""

    title = "Valid Repositories"
    columns = [
        RepositoryGrid.NameColumn("Name", key="name", attach_popup=True),
        RepositoryGrid.DescriptionColumn("Synopsis", key="description", attach_popup=False),
        RepositoryGrid.TypeColumn("Type"),
        InstallableRevisionColumn("Installable Revisions"),
        RepositoryGrid.UserColumn("Owner", model_class=model.User, attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        RepositoryCategoryColumn("Category", model_class=model.Category, key="Category.name", visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name, description",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )

    def build_initial_query(self, trans, **kwd):
        filter = trans.app.repository_grid_filter_manager.get_filter(trans)
        if "id" in kwd:
            # The user is browsing categories of valid repositories, so filter the request by the received id,
            # which is a category id.
            if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
                return (
                    trans.sa_session.query(model.Repository)
                    .join(model.RepositoryMetadata.table)
                    .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                    .join(model.User.table)
                    .join(model.RepositoryCategoryAssociation)
                    .join(model.Category.table)
                    .filter(
                        and_(
                            model.Category.__table__.c.id == trans.security.decode_id(kwd["id"]),
                            model.RepositoryMetadata.table.c.downloadable == true(),
                        )
                    )
                )
            if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
                return (
                    trans.sa_session.query(model.Repository)
                    .filter(model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION)
                    .join(model.RepositoryMetadata.table)
                    .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                    .join(model.User.table)
                    .join(model.RepositoryCategoryAssociation)
                    .join(model.Category.table)
                    .filter(
                        and_(
                            model.Category.table.c.id == trans.security.decode_id(kwd["id"]),
                            model.RepositoryMetadata.table.c.downloadable == true(),
                        )
                    )
                )
            else:
                # The value of filter is None.
                return (
                    trans.sa_session.query(model.Repository)
                    .filter(
                        and_(
                            model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false()
                        )
                    )
                    .join(model.RepositoryMetadata.table)
                    .join(model.User.table)
                    .join(model.RepositoryCategoryAssociation)
                    .join(model.Category.table)
                    .filter(
                        and_(
                            model.Category.table.c.id == trans.security.decode_id(kwd["id"]),
                            model.RepositoryMetadata.table.c.downloadable == true(),
                        )
                    )
                )
        # The user performed a free text search on the ValidCategoryGrid.
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE:
            return (
                trans.sa_session.query(model.Repository)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
                .filter(model.RepositoryMetadata.table.c.downloadable == true())
            )
        if filter == trans.app.repository_grid_filter_manager.filters.CERTIFIED_LEVEL_ONE_SUITES:
            return (
                trans.sa_session.query(model.Repository)
                .filter(model.Repository.type == rt_util.REPOSITORY_SUITE_DEFINITION)
                .join(model.RepositoryMetadata.table)
                .filter(or_(*trans.app.repository_registry.certified_level_one_clause_list))
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
                .filter(model.RepositoryMetadata.table.c.downloadable == true())
            )
        else:
            # The value of filter is None.
            return (
                trans.sa_session.query(model.Repository)
                .filter(
                    and_(model.Repository.table.c.deleted == false(), model.Repository.table.c.deprecated == false())
                )
                .join(model.RepositoryMetadata.table)
                .join(model.User.table)
                .outerjoin(model.RepositoryCategoryAssociation)
                .outerjoin(model.Category.table)
                .filter(model.RepositoryMetadata.table.c.downloadable == true())
            )



"""
Details of how the data model objects are mapped onto the relational database
are encapsulated here.
"""

import logging
from typing import (
    Any,
    Optional,
    TYPE_CHECKING,
)

import tool_shed.webapp.model
import tool_shed.webapp.util.shed_statistics as shed_statistics
from galaxy.model.base import SharedModelMapping
from galaxy.model.orm.engine_factory import build_engine
from tool_shed.webapp.model import mapper_registry
from tool_shed.webapp.security import CommunityRBACAgent

if TYPE_CHECKING:
    from tool_shed.webapp.model import User as ToolShedUser

log = logging.getLogger(__name__)

metadata = mapper_registry.metadata


class ToolShedModelMapping(SharedModelMapping):
    User: type["ToolShedUser"]
    security_agent: CommunityRBACAgent
    shed_counter: shed_statistics.ShedCounter
    create_tables: bool


def init(
    url: str, engine_options: Optional[dict[str, Any]] = None, create_tables: bool = False
) -> ToolShedModelMapping:
    """Connect mappings to the database"""
    engine_options = engine_options or {}
    # Create the database engine
    engine = build_engine(url, engine_options)

    result = ToolShedModelMapping([tool_shed.webapp.model], engine=engine)

    if create_tables:
        metadata.create_all(bind=engine)

    result.create_tables = create_tables

    result.security_agent = CommunityRBACAgent(result)
    result.shed_counter = shed_statistics.ShedCounter(result)

    session = result.session()
    with session.begin():
        result.shed_counter.generate_statistics(session)

    return result



import logging

import galaxy.tools
from galaxy.tool_shed.util.tool_util import (
    copy_sample_file,
    copy_sample_files,
    generate_message_for_invalid_tools,
)
from galaxy.util.expressions import ExpressionContext

log = logging.getLogger(__name__)


def new_state(trans, tool, invalid=False):
    """Create a new `DefaultToolState` for the received tool.  Only inputs on the first page will be initialized."""
    state = galaxy.tools.DefaultToolState()
    state.inputs = {}
    if invalid:
        # We're attempting to display a tool in the tool shed that has been determined to have errors, so is invalid.
        return state
    try:
        # Attempt to generate the tool state using the standard Galaxy-side code
        return tool.new_state(trans)
    except Exception as e:
        # Fall back to building tool state as below
        log.debug(
            'Failed to build tool state for tool "%s" using standard method, will try to fall back on custom method: %s',
            tool.id,
            e,
        )
    inputs = tool.inputs_by_page[0]
    context = ExpressionContext(state.inputs, parent=None)
    for input in inputs.values():
        try:
            state.inputs[input.name] = input.get_initial_value(trans, context)
        except Exception:
            # FIXME: not all values should be an empty list
            state.inputs[input.name] = []
    return state


__all__ = (
    "copy_sample_file",
    "copy_sample_files",
    "generate_message_for_invalid_tools",
    "new_state",
)



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_maker_repository_name = "column_maker_0110"
column_maker_repository_description = "A flexible aligner."
column_maker_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

emboss_repository_name = "emboss_0110"
emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools"

category_name = "Test 0110 Invalid Repository Dependencies"
category_desc = "Test 0110 Invalid Repository Dependencies"


class TestBasicRepositoryDependencies(ShedTwillTestCase):
    """Testing emboss 5 with repository dependencies."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_category(self):
        """Create a category for this test suite"""
        self.create_category(name=category_name, description=category_desc)

    def test_0010_create_emboss_datatypes_repository_and_upload_tarball(self):
        """Create and populate the emboss_datatypes repository."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        category = self.populator.get_category_with_name(category_name)
        column_maker_repository = self.get_or_create_repository(
            name=column_maker_repository_name,
            description=column_maker_repository_description,
            long_description=column_maker_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            column_maker_repository,
            "column_maker/column_maker.tar",
            commit_message="Uploaded column_maker tarball.",
        )

    def test_0020_create_emboss_5_repository_and_upload_files(self):
        """Create and populate the emboss_5_0110 repository."""
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "emboss/emboss.tar",
            commit_message="Uploaded emboss tool tarball.",
        )

    def test_0025_generate_repository_dependency_with_invalid_url(self):
        """Generate a repository dependency for emboss 5 with an invalid URL."""
        dependency_path = self.generate_temp_path("test_0110", additional_paths=["simple"])
        column_maker_repository = self._get_repository_by_name_and_owner(
            column_maker_repository_name, common.test_user_1_name
        )
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        url = "http://http://this is not an url!"
        name = column_maker_repository.name
        owner = column_maker_repository.owner
        changeset_revision = self.get_repository_tip(column_maker_repository)
        strings_displayed = ["Repository dependencies are currently supported only within the same tool shed"]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=emboss_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=False,
        )

    def test_0030_generate_repository_dependency_with_invalid_name(self):
        """Generate a repository dependency for emboss 5 with an invalid name."""
        dependency_path = self.generate_temp_path("test_0110", additional_paths=["simple"])
        repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        url = self.url
        name = "!?invalid?!"
        owner = repository.owner
        changeset_revision = self.get_repository_tip(repository)
        strings_displayed = ["because the name is invalid."]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=emboss_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=False,
        )

    def test_0035_generate_repository_dependency_with_invalid_owner(self):
        """Generate a repository dependency for emboss 5 with an invalid owner."""
        dependency_path = self.generate_temp_path("test_0110", additional_paths=["simple"])
        repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        url = self.url
        name = repository.name
        owner = "!?invalid?!"
        changeset_revision = self.get_repository_tip(repository)
        strings_displayed = ["because the owner is invalid."]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=emboss_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=False,
        )

    def test_0040_generate_repository_dependency_with_invalid_changeset_revision(self):
        """Generate a repository dependency for emboss 5 with an invalid changeset revision."""
        dependency_path = self.generate_temp_path("test_0110", additional_paths=["simple", "invalid"])
        repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        url = self.url
        name = repository.name
        owner = repository.owner
        changeset_revision = "!?invalid?!"
        strings_displayed = ["because the changeset revision is invalid."]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=emboss_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=False,
        )



from sqlalchemy import select

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.security.validate_user_input import (
    validate_email,
    validate_password,
    validate_publicname,
)
from tool_shed.context import ProvidesUserContext
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp.model import User
from tool_shed_client.schema import (
    CreateUserRequest,
    UserV2 as ApiUser,
)


def index(app: ToolShedApp, deleted: bool) -> list[ApiUser]:
    users: list[ApiUser] = []
    for user in get_users_by_deleted(app.model.context, User, deleted):
        users.append(get_api_user(app, user))
    return users


def create_user(app: ToolShedApp, email: str, username: str, password: str) -> User:
    if username == "repos":
        raise RequestParameterInvalidException("Cannot create a tool shed user with the username repos")
    sa_session = app.model.context
    user = User(email=email)
    user.set_password_cleartext(password)
    user.username = username
    # API was doing this but mypy doesn't think user has an active boolean attribute.
    # if app.config.user_activation_on:
    #    user.active = False
    # else:
    #    user.active = True  # Activation is off, every new user is active by default.
    sa_session.add(user)
    sa_session.commit()
    app.security_agent.create_private_user_role(user)
    return user


def api_create_user(trans: ProvidesUserContext, request: CreateUserRequest) -> ApiUser:
    app = trans.app
    message = _validate(
        trans, email=request.email, password=request.password, confirm=request.password, username=request.username
    )
    if message:
        raise RequestParameterInvalidException(message)
    user = create_user(app, request.email, request.username, request.password)
    return get_api_user(app, user)


def get_api_user(app: ToolShedApp, user: User) -> ApiUser:
    admin = app.config.is_admin_user(user)
    return ApiUser(
        id=app.security.encode_id(user.id),
        username=user.username,
        is_admin=admin,
    )


def _validate(trans: ProvidesUserContext, email: str, password: str, confirm: str, username: str) -> str:
    if username in ["repos"]:
        return f"The term '{username}' is a reserved word in the Tool Shed, so it cannot be used as a public user name."
    message = "\n".join(
        (
            validate_email(trans, email),
            validate_password(trans, password, confirm),
            validate_publicname(trans, username),
        )
    ).rstrip()
    return message


def get_users_by_deleted(session, user_model, deleted):
    stmt = select(user_model).where(user_model.deleted == deleted).order_by(user_model.username)
    return session.scalars(stmt)



from typing import TYPE_CHECKING

from sqlalchemy import and_
from sqlalchemy.orm import joinedload

from tool_shed.webapp.model import (
    Repository,
    User,
)

if TYPE_CHECKING:
    from sqlalchemy.orm import scoped_session


def get_repository_query(session: "scoped_session"):
    return session.query(Repository)


def get_repository_by_name(session: "scoped_session", name):
    """Get a repository from the database via name."""
    return get_repository_query(session).filter_by(name=name).first()


def get_repository_by_name_and_owner(session: "scoped_session", name, owner, eagerload_columns=None):
    """Get a repository from the database via name and owner"""
    repository_query = get_repository_query(session)
    q = repository_query.filter(
        and_(
            Repository.name == name,
            User.username == owner,
            Repository.user_id == User.id,
        )
    )
    if eagerload_columns:
        q = q.options(joinedload(*eagerload_columns))
    return q.first()



import logging
import time

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repository_name = "filtering_1410"
repository_description = "Galaxy's filtering tool"
repository_long_description = "Long description of Galaxy's filtering repository"

category_name = "Test 1410 - Galaxy Update Manager"
category_description = "Functional test suite to test the update manager."

"""
1. Create and populate the filtering_1410 repository.
2. Install filtering_1410 to Galaxy.
3. Upload a readme file.
4. Verify that the browse page now shows an update available.
"""


class TestUpdateManager(ShedTwillTestCase):
    """Test the Galaxy update manager."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user.

        Create all the user accounts that are needed for this test script to run independently of other tests.
        Previously created accounts will not be re-created.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_filtering_repository(self):
        """Create and populate the filtering_1410 repository.

        We are at step 1 - Create and populate the filtering_1410 repository.
        Create filtering_1410 and upload the tool tarball to it.
        """
        self.login(email=common.admin_email, username=common.admin_username)
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
        )
        self.commit_tar_to_repository(
            repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Uploaded filtering 1.1.0",
        )

    def test_0010_install_filtering_repository(self):
        """Install the filtering_1410 repository.

        We are at step 2 - Install filtering_1410 to Galaxy.
        Install the filtering repository to Galaxy.
        """
        self._install_repository(
            repository_name, common.test_user_1_name, category_name, new_tool_panel_section_label="test_1410"
        )
        installed_repository = self._get_installed_repository_by_name_owner(repository_name, common.test_user_1_name)
        self._assert_has_installed_repos_with_names(repository_name)
        self._assert_has_valid_tool_with_name("Filter")
        self._assert_repo_has_tool_with_id(installed_repository, "Filter1")

    def test_0015_upload_readme_file(self):
        """Upload readme.txt to filtering_1410.

        We are at step 3 - Upload a readme file.
        Upload readme.txt. This will have the effect of making the installed changeset revision not be the most recent downloadable revision,
        but without generating a second downloadable revision. Then sleep for 3 seconds to make sure the update manager picks up the new
        revision.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "readme.txt")

    def test_0020_check_for_displayed_update(self):
        """Browse installed repositories and verify update.

        We are at step 4 - Verify that the browse page now shows an update available.
        The browse page should now show filtering_1410 as installed, but with a yellow box indicating that there is an update available.
        """
        # Wait 3 seconds, just to be sure we're past hours_between_check.
        time.sleep(3)
        installed_repository = self._get_installed_repository_by_name_owner(repository_name, common.test_user_1_name)
        response = self.update_installed_repository(installed_repository)
        assert response["status"] == "ok"
        assert "has been updated" in response["message"]



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

category_name = "Test 1460 Data Manager"
category_description = "Test script 1460 for testing Data Managers"
data_manager_repository_name = "data_manager_1460"
data_manager_repository_description = "Repository that contains a Data Manager"
data_manager_repository_long_description = f"{data_manager_repository_name}: {data_manager_repository_description}"
data_manager_name = "testing_data_manager"
data_manager_tar_file = "1460_files/data_manager_files/test_data_manager.tar"

"""
1. Add a Data Manager to toolshed

2. install Data Manager

3. Check that Data Manager tool

"""

# TODO: Allow testing actual Execution of installed Data Manager Tool.


class TestDataManagers(ShedTwillTestCase):
    """Test installing a repository containing a Data Manager."""

    requires_galaxy = True

    def test_0000_initiate_users_and_category(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.admin_email, username=common.admin_username)
        self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)

    def test_0010_create_data_manager_repository(self):
        """Create and populate data_manager_1460.

        This is step 1 - Create repository data_manager_1460.

        Create and populate a repository that contains a Data manager.
        """
        category = self.populator.get_category_with_name(category_name)
        repository = self.get_or_create_repository(
            name=data_manager_repository_name,
            description=data_manager_repository_description,
            long_description=data_manager_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        # Upload the data manager files to the repository.
        self.commit_tar_to_repository(
            repository,
            data_manager_tar_file,
            commit_message=f"Populate {data_manager_repository_name} with a data manager configuration.",
        )

    def test_0020_install_data_manager_repository(self):
        """Install the data_manager_1460 repository to galaxy.

        This is step 3 - Attempt to install the repository into a galaxy instance, verify that it is installed.
        """
        self._install_repository(
            data_manager_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
        )

    def test_0030_verify_data_manager_tool(self):
        """Verify that the data_manager_1460 repository is installed and Data Manager tool appears in list in Galaxy."""
        repository = self._get_installed_repository_by_name_owner(data_manager_repository_name, common.test_user_1_name)
        if self.full_stack_galaxy:
            strings_displayed = ["status", "jobs", data_manager_name]
            self.display_installed_jobs_list_page(
                repository, data_manager_names=data_manager_name, strings_displayed=strings_displayed
            )

    def test_0040_verify_data_manager_data_table(self):
        """Verify that the installed repository populated shed_tool_data_table.xml and the sample files."""
        self.verify_installed_repository_data_table_entries(
            required_data_table_entries=["data_manager_test_data_table"]
        )



import logging

import tool_shed.grids.admin_grids as admin_grids
from galaxy import (
    util,
    web,
)
from galaxy.util import inflector
from galaxy.web.legacy_framework import grids
from galaxy.webapps.base.controller import BaseUIController
from tool_shed.metadata import repository_metadata_manager
from tool_shed.util import (
    metadata_util,
    repository_util,
    shed_util_common as suc,
)
from tool_shed.util.admin_util import Admin
from tool_shed.util.web_util import escape
from tool_shed.webapp.model import Category

log = logging.getLogger(__name__)


class AdminController(BaseUIController, Admin):
    user_list_grid = admin_grids.UserGrid()
    role_list_grid = admin_grids.RoleGrid()
    group_list_grid = admin_grids.GroupGrid()
    manage_category_grid = admin_grids.ManageCategoryGrid()
    repository_grid = admin_grids.AdminRepositoryGrid()
    repository_metadata_grid = admin_grids.RepositoryMetadataGrid()

    delete_operation = grids.GridOperation("Delete", condition=(lambda item: not item.deleted), allow_multiple=True)
    undelete_operation = grids.GridOperation(
        "Undelete", condition=(lambda item: item.deleted and not item.purged), allow_multiple=True
    )
    purge_operation = grids.GridOperation(
        "Purge", condition=(lambda item: item.deleted and not item.purged), allow_multiple=True
    )

    @web.expose
    @web.require_admin
    def browse_repositories(self, trans, **kwd):
        # We add parameters to the keyword dict in this method in order to rename the param
        # with an "f-" prefix, simulating filtering by clicking a search link.  We have
        # to take this approach because the "-" character is illegal in HTTP requests.
        if "operation" in kwd:
            operation = kwd["operation"].lower()
            if operation == "view_or_manage_repository":
                return trans.response.send_redirect(
                    web.url_for(controller="repository", action="browse_repositories", **kwd)
                )
            elif operation == "edit_repository":
                return trans.response.send_redirect(
                    web.url_for(controller="repository", action="edit_repository", **kwd)
                )
            elif operation == "repositories_by_user":
                # Eliminate the current filters if any exist.
                for k in list(kwd.keys()):
                    if k.startswith("f-"):
                        del kwd[k]
                if "user_id" in kwd:
                    user = suc.get_user(trans.app, kwd["user_id"])
                    kwd["f-email"] = user.email
                    del kwd["user_id"]
                else:
                    # The received id is the repository id, so we need to get the id of the user
                    # that uploaded the repository.
                    repository_id = kwd.get("id", None)
                    repository = repository_util.get_repository_in_tool_shed(trans.app, repository_id)
                    kwd["f-email"] = repository.user.email
            elif operation == "repositories_by_category":
                # Eliminate the current filters if any exist.
                for k in list(kwd.keys()):
                    if k.startswith("f-"):
                        del kwd[k]
                category_id = kwd.get("id")
                assert category_id
                category = suc.get_category(trans.app, category_id)
                kwd["f-Category.name"] = category.name
            elif operation == "receive email alerts":
                if kwd["id"]:
                    kwd["caller"] = "browse_repositories"
                    return trans.response.send_redirect(
                        web.url_for(controller="repository", action="set_email_alerts", **kwd)
                    )
                else:
                    del kwd["operation"]
            elif operation == "delete":
                return self.delete_repository(trans, **kwd)
            elif operation == "undelete":
                return self.undelete_repository(trans, **kwd)
        # The changeset_revision_select_field in the RepositoryGrid performs a refresh_on_change
        # which sends in request parameters like changeset_revison_1, changeset_revision_2, etc.  One
        # of the many select fields on the grid performed the refresh_on_change, so we loop through
        # all of the received values to see which value is not the repository tip.  If we find it, we
        # know the refresh_on_change occurred, and we have the necessary repository id and change set
        # revision to pass on.
        for k, v in kwd.items():
            changeset_revision_str = "changeset_revision_"
            if k.startswith(changeset_revision_str):
                repository_id = trans.security.encode_id(int(k.lstrip(changeset_revision_str)))
                repository = repository_util.get_repository_in_tool_shed(trans.app, repository_id)
                if repository.tip() != v:
                    return trans.response.send_redirect(
                        web.url_for(
                            controller="repository",
                            action="browse_repositories",
                            operation="view_or_manage_repository",
                            id=trans.security.encode_id(repository.id),
                            changeset_revision=v,
                        )
                    )
        # Render the list view
        return self.repository_grid(trans, **kwd)

    @web.expose
    @web.require_admin
    def browse_repository_metadata(self, trans, **kwd):
        if "operation" in kwd:
            operation = kwd["operation"].lower()
            if operation == "delete":
                return self.delete_repository_metadata(trans, **kwd)
            if operation == "view_or_manage_repository_revision":
                # The received id is a RepositoryMetadata object id, so we need to get the
                # associated Repository and redirect to view_or_manage_repository with the
                # changeset_revision.
                repository_metadata = metadata_util.get_repository_metadata_by_id(trans.app, kwd["id"])
                repository = repository_metadata.repository
                kwd["id"] = trans.security.encode_id(repository.id)
                kwd["changeset_revision"] = repository_metadata.changeset_revision
                kwd["operation"] = "view_or_manage_repository"
                return trans.response.send_redirect(
                    web.url_for(controller="repository", action="browse_repositories", **kwd)
                )
        return self.repository_metadata_grid(trans, **kwd)

    @web.expose
    @web.require_admin
    def create_category(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        name = kwd.get("name", "").strip()
        description = kwd.get("description", "").strip()
        if kwd.get("create_category_button", False):
            if not name or not description:
                message = "Enter a valid name and a description"
                status = "error"
            elif suc.get_category_by_name(trans.app, name):
                message = "A category with that name already exists"
                status = "error"
            else:
                # Create the category
                category = Category(name=name, description=description)
                trans.sa_session.add(category)
                trans.sa_session.commit()
                # Update the Tool Shed's repository registry.
                trans.app.repository_registry.add_category_entry(category)
                message = f"Category '{escape(category.name)}' has been created"
                status = "done"
                trans.response.send_redirect(
                    web.url_for(controller="admin", action="manage_categories", message=message, status=status)
                )
        return trans.fill_template(
            "/webapps/tool_shed/category/create_category.mako",
            name=name,
            description=description,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def delete_repository(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        if id := kwd.get("id", None):
            # Deleting multiple items is currently not allowed (allow_multiple=False), so there will only be 1 id.
            ids = util.listify(id)
            deleted_repositories = []
            for repository_id in ids:
                repository = repository_util.get_repository_in_tool_shed(trans.app, repository_id)
                if repository:
                    if not repository.deleted:
                        # Mark all installable repository_metadata records as not installable.
                        for repository_metadata in repository.downloadable_revisions:
                            repository_metadata.downloadable = False
                            trans.sa_session.add(repository_metadata)
                        # Mark the repository admin role as deleted.
                        repository_admin_role = repository.admin_role
                        if repository_admin_role is not None:
                            repository_admin_role.deleted = True
                            trans.sa_session.add(repository_admin_role)
                        repository.deleted = True
                        trans.sa_session.add(repository)
                        trans.sa_session.commit()
                        # Update the repository registry.
                        trans.app.repository_registry.remove_entry(repository)
                        deleted_repositories.append(repository.name)
            count = len(deleted_repositories)
            if count:
                message = "Deleted {} {}: {}".format(
                    count,
                    inflector.cond_plural(count, "repository"),
                    escape(" ".join(deleted_repositories)),
                )
            else:
                message = "All selected repositories were already marked deleted."
        else:
            message = "No repository ids received for deleting."
            status = "error"
        trans.response.send_redirect(
            web.url_for(
                controller="admin", action="browse_repositories", message=util.sanitize_text(message), status=status
            )
        )

    @web.expose
    @web.require_admin
    def delete_repository_metadata(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        if id := kwd.get("id", None):
            ids = util.listify(id)
            for repository_metadata_id in ids:
                repository_metadata = metadata_util.get_repository_metadata_by_id(trans.app, repository_metadata_id)
                trans.sa_session.delete(repository_metadata)
                trans.sa_session.commit()
            count = len(ids)
            message = "Deleted {} repository metadata {}".format(count, inflector.cond_plural(count, "record"))
        else:
            message = "No repository metadata ids received for deleting."
            status = "error"
        trans.response.send_redirect(
            web.url_for(
                controller="admin",
                action="browse_repository_metadata",
                message=util.sanitize_text(message),
                status=status,
            )
        )

    @web.expose
    @web.require_admin
    def edit_category(self, trans, **kwd):
        """Handle requests to edit TS category name or description"""
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        id = kwd.get("id")
        if not id:
            message = "No category ids received for editing"
            trans.response.send_redirect(
                web.url_for(controller="admin", action="manage_categories", message=message, status="error")
            )
        assert id is not None  # redundant, can be dropped when we annotate trans
        category = suc.get_category(trans.app, id)
        original_category_name = str(category.name)
        original_category_description = str(category.description)
        if kwd.get("edit_category_button", False):
            flush_needed = False
            new_name = kwd.get("name", "").strip()
            new_description = kwd.get("description", "").strip()
            if original_category_name != new_name:
                if not new_name:
                    message = "Enter a valid name"
                    status = "error"
                elif original_category_name != new_name and suc.get_category_by_name(trans.app, new_name):
                    message = "A category with that name already exists"
                    status = "error"
                else:
                    category.name = new_name
                    flush_needed = True
            if original_category_description != new_description:
                category.description = new_description
                if not flush_needed:
                    flush_needed = True
            if flush_needed:
                trans.sa_session.add(category)
                trans.sa_session.commit()
                if original_category_name != new_name:
                    # Update the Tool Shed's repository registry.
                    trans.app.repository_registry.edit_category_entry(original_category_name, new_name)
                message = f"The information has been saved for category '{escape(category.name)}'"
                status = "done"
                return trans.response.send_redirect(
                    web.url_for(controller="admin", action="manage_categories", message=message, status=status)
                )
        return trans.fill_template(
            "/webapps/tool_shed/category/edit_category.mako", category=category, message=message, status=status
        )

    @web.expose
    @web.require_admin
    def manage_categories(self, trans, **kwd):
        if "f-free-text-search" in kwd:
            # Trick to enable searching repository name, description from the CategoryGrid.
            # What we've done is rendered the search box for the RepositoryGrid on the grid.mako
            # template for the CategoryGrid.  See ~/templates/webapps/tool_shed/category/grid.mako.
            # Since we are searching repositories and not categories, redirect to browse_repositories().
            return trans.response.send_redirect(web.url_for(controller="admin", action="browse_repositories", **kwd))
        if "operation" in kwd:
            operation = kwd["operation"].lower()
            if operation == "create":
                return trans.response.send_redirect(web.url_for(controller="admin", action="create_category", **kwd))
            elif operation == "delete":
                return trans.response.send_redirect(
                    web.url_for(controller="admin", action="mark_category_deleted", **kwd)
                )
            elif operation == "undelete":
                return trans.response.send_redirect(web.url_for(controller="admin", action="undelete_category", **kwd))
            elif operation == "purge":
                return trans.response.send_redirect(web.url_for(controller="admin", action="purge_category", **kwd))
            elif operation == "edit":
                return trans.response.send_redirect(web.url_for(controller="admin", action="edit_category", **kwd))
        return self.manage_category_grid(trans, **kwd)

    @web.expose
    @web.require_admin
    def regenerate_statistics(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        status = kwd.get("status", "done")
        if "regenerate_statistics_button" in kwd:
            trans.app.shed_counter.generate_statistics(trans.sa_session)
            message = "Successfully regenerated statistics"
        return trans.fill_template("/webapps/tool_shed/admin/statistics.mako", message=message, status=status)

    @web.expose
    @web.require_admin
    def manage_role_associations(self, trans, **kwd):
        """Manage users, groups and repositories associated with a role."""
        role_id = kwd.get("id", None)
        role = repository_util.get_role_by_id(trans.app, role_id)
        # We currently only have a single role associated with a repository, the repository admin role.
        repository_role_association = role.repositories[0]
        repository = repository_role_association.repository
        associations_dict = repository_util.handle_role_associations(trans.app, role, repository, **kwd)
        in_users = associations_dict.get("in_users", [])
        out_users = associations_dict.get("out_users", [])
        in_groups = associations_dict.get("in_groups", [])
        out_groups = associations_dict.get("out_groups", [])
        message = associations_dict.get("message", "")
        status = associations_dict.get("status", "done")
        return trans.fill_template(
            "/webapps/tool_shed/role/role.mako",
            in_admin_controller=True,
            repository=repository,
            role=role,
            in_users=in_users,
            out_users=out_users,
            in_groups=in_groups,
            out_groups=out_groups,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def reset_metadata_on_selected_repositories_in_tool_shed(self, trans, **kwd):
        rmm = repository_metadata_manager.RepositoryMetadataManager(trans)
        if "reset_metadata_on_selected_repositories_button" in kwd:
            message, status = rmm.reset_metadata_on_selected_repositories(**kwd)
        else:
            message = escape(util.restore_text(kwd.get("message", "")))
            status = kwd.get("status", "done")
        repositories_select_field = rmm.build_repository_ids_select_field(
            name="repository_ids", multiple=True, display="checkboxes", my_writable=False
        )
        return trans.fill_template(
            "/webapps/tool_shed/common/reset_metadata_on_selected_repositories.mako",
            repositories_select_field=repositories_select_field,
            message=message,
            status=status,
        )

    @web.expose
    @web.require_admin
    def undelete_repository(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        if id := kwd.get("id", None):
            # Undeleting multiple items is currently not allowed (allow_multiple=False), so there will only be 1 id.
            ids = util.listify(id)
            undeleted_repositories = []
            for repository_id in ids:
                repository = repository_util.get_repository_in_tool_shed(trans.app, repository_id)
                if repository:
                    if repository.deleted:
                        # Inspect all repository_metadata records to determine those that are installable, and mark
                        # them accordingly.
                        for repository_metadata in repository.metadata_revisions:
                            metadata = repository_metadata.metadata
                            if metadata:
                                if metadata_util.is_downloadable(metadata):
                                    repository_metadata.downloadable = True
                                    trans.sa_session.add(repository_metadata)
                        # Mark the repository admin role as not deleted.
                        repository_admin_role = repository.admin_role
                        if repository_admin_role is not None:
                            repository_admin_role.deleted = False
                            trans.sa_session.add(repository_admin_role)
                        repository.deleted = False
                        trans.sa_session.add(repository)
                        trans.sa_session.commit()
                        if not repository.deprecated:
                            # Update the repository registry.
                            trans.app.repository_registry.add_entry(repository)
                        undeleted_repositories.append(repository.name)
            count = len(undeleted_repositories)
            if count:
                message = "Undeleted {} {}: {}".format(
                    count,
                    inflector.cond_plural(count, "repository"),
                    " ".join(undeleted_repositories),
                )
            else:
                message = "No selected repositories were marked deleted, so they could not be undeleted."
        else:
            message = "No repository ids received for undeleting."
        trans.response.send_redirect(
            web.url_for(
                controller="admin", action="browse_repositories", message=util.sanitize_text(message), status="done"
            )
        )

    @web.expose
    @web.require_admin
    def mark_category_deleted(self, trans, **kwd):
        # TODO: We should probably eliminate the Category.deleted column since it really makes no
        # sense to mark a category as deleted (category names and descriptions can be changed instead).
        # If we do this, and the following 2 methods can be eliminated.
        if id := kwd.get("id", None):
            ids = util.listify(id)
            deleted_categories = []
            for category_id in ids:
                category = suc.get_category(trans.app, category_id)
                category.deleted = True
                trans.sa_session.add(category)
                trans.sa_session.commit()
                # Update the Tool Shed's repository registry.
                trans.app.repository_registry.remove_category_entry(category)
                deleted_categories.append(category.name)
            message = "Deleted {} categories: {}".format(len(deleted_categories), escape(" ".join(deleted_categories)))
        else:
            message = "No category ids received for deleting."
        trans.response.send_redirect(
            web.url_for(
                controller="admin", action="manage_categories", message=util.sanitize_text(message), status="done"
            )
        )

    @web.expose
    @web.require_admin
    def purge_category(self, trans, **kwd):
        # This method should only be called for a Category that has previously been deleted.
        # Purging a deleted Category deletes all of the following from the database:
        # - RepoitoryCategoryAssociations where category_id == Category.id
        message = escape(kwd.get("message", ""))
        if id := kwd.get("id", None):
            ids = util.listify(id)
            purged_categories = []
            for category_id in ids:
                category = suc.get_category(trans.app, category_id)
                if category.deleted:
                    # Delete RepositoryCategoryAssociations
                    for rca in category.repositories:
                        trans.sa_session.delete(rca)
                    trans.sa_session.commit()
                    purged_categories.append(category.name)
            message = "Purged {} categories: {}".format(len(purged_categories), escape(" ".join(purged_categories)))
        else:
            message = "No category ids received for purging."
        trans.response.send_redirect(
            web.url_for(
                controller="admin", action="manage_categories", message=util.sanitize_text(message), status="done"
            )
        )

    @web.expose
    @web.require_admin
    def undelete_category(self, trans, **kwd):
        message = escape(kwd.get("message", ""))
        if id := kwd.get("id", None):
            ids = util.listify(id)
            undeleted_categories = []
            for category_id in ids:
                category = suc.get_category(trans.app, category_id)
                if category.deleted:
                    category.deleted = False
                    trans.sa_session.add(category)
                    trans.sa_session.commit()
                    # Update the Tool Shed's repository registry.
                    trans.app.repository_registry.add_category_entry(category)
                    undeleted_categories.append(category.name)
            message = "Undeleted {} categories: {}".format(
                len(undeleted_categories), escape(" ".join(undeleted_categories))
            )
        else:
            message = "No category ids received for undeleting."
        trans.response.send_redirect(
            web.url_for(
                controller="admin", action="manage_categories", message=util.sanitize_text(message), status="done"
            )
        )



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repository_name = "filtering_0420"
repository_description = "Galaxy filtering tool for test 0420"
repository_long_description = "Long description of Galaxy filtering tool for test 0420"

first_changeset_hash = ""

"""
1. Add and populate a repository to the tool shed with change set revision 0 (assume owner is test).
2. Add valid change set revision 1.
3. Visit the following url and check for appropriate strings: <tool shed base url>/view/user1
4. Visit the following url and check for appropriate strings: <tool shed base url>/view/user1/filtering_0420 .
   The resulting page should contain change set revision 1
5. Visit the following url and check for appropriate strings: <tool shed base url>/view/user1/filtering_0420/<revision 0> .
   The resulting page should not contain change set revision 1, but should contain change set revision 0.
6. Visit the following url and check for appropriate strings: <tool shed base url>/view/user1/filtering_0420/<invalid revision>
7. Visit the following url and check for appropriate strings: <tool shed base url>/view/user1/<invalid repository name>
8. Visit the following url and check for appropriate strings: <tool shed base url>/view/<invalid owner>
"""


class TestRepositoryCitableURLs(ShedTwillTestCase):
    """Test repository citable url features."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user.

        Create all the user accounts that are needed for this test script to run independently of other tests.
        Previously created accounts will not be re-created.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username, explicit_logout=True)

    def test_0005_create_repository(self):
        """Create and populate the filtering_0420 repository

        We are at step 1.
        Add and populate a repository to the tool shed with change set revision 0 (assume owner is test_user_1).
        """
        global first_changeset_hash
        category = self.create_category(
            name="Test 0400 Repository Citable URLs", description="Test 0400 Repository Citable URLs category"
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        strings_displayed = self.expect_repo_created_strings(repository_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=strings_displayed,
        )
        self.commit_tar_to_repository(
            repository,
            "filtering/filtering_2.2.0.tar",
            commit_message="Uploaded filtering 2.2.0 tarball.",
        )
        # We'll be checking for this hash later, after uploading another file to the repository, making get_repository_tip() not usable.
        first_changeset_hash = self.get_repository_tip(repository)

    def test_0010_upload_new_file_to_repository(self):
        """Upload a readme file to the repository in order to create a second changeset revision.

        We are at step 2.
        Add valid change set revision 1.
        The repository should now contain two changeset revisions, 0:<revision hash> and 1:<revision hash>.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "readme.txt")

    def test_0015_load_user_view_page(self):
        """Load the /view/<username> page amd check for strings.

        We are at step 3.
        Visit the following url and check for appropriate strings: <tool shed base url>/view/user1
        """
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        encoded_user_id = self.security.encode_id(test_user_1.id)
        # Since twill does not load the contents of an iframe, we need to check that the iframe has been generated correctly,
        # then directly load the url that the iframe should be loading and check for the expected strings.
        # The iframe should point to /repository/browse_repositories?user_id=<encoded user ID>&operation=repositories_by_user
        if self.is_v2:
            strings_displayed = []
        else:
            strings_displayed = ["/repository/browse_repositories", encoded_user_id, "operation=repositories_by_user"]
            strings_displayed.append(encoded_user_id)
        strings_displayed_in_iframe = ["user1", "filtering_0420", repository_description]
        self.load_citable_url(
            username="user1",
            repository_name=None,
            changeset_revision=None,
            encoded_user_id=encoded_user_id,
            encoded_repository_id=None,
            strings_displayed=strings_displayed,
            strings_displayed_in_iframe=strings_displayed_in_iframe,
        )

    def test_0020_load_repository_view_page(self):
        """Load the /view/<user>/<repository> page and check for the appropriate strings.

        We are at step 4.
        Visit the following url and check for strings: ``<tool shed base url>/view/user1/filtering_0420`` .
        The resulting page should contain change set revision 1
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        encoded_user_id = self.security.encode_id(test_user_1.id)
        encoded_repository_id = repository.id
        # Since twill does not load the contents of an iframe, we need to check that the iframe has been generated correctly,
        # then directly load the url that the iframe should be loading and check for the expected strings.
        # The iframe should point to /repository/bview_repository?id=<encoded repository ID>
        if self.is_v2:
            strings_displayed = []
        else:
            strings_displayed = ["/repository", "view_repository", "id=", encoded_repository_id]
        strings_displayed_in_iframe = [
            "user1",
            "filtering_0420",
            self._escape_page_content_if_needed(repository_long_description),
        ]
        strings_displayed_in_iframe.append(self.get_repository_tip(repository))
        if not self.is_v2:
            strings_displayed_in_iframe.append("Link to this repository:")
            strings_displayed_in_iframe.append(f"{self.url}/view/user1/filtering_0420")
        self.load_citable_url(
            username="user1",
            repository_name="filtering_0420",
            changeset_revision=None,
            encoded_user_id=encoded_user_id,
            encoded_repository_id=encoded_repository_id,
            strings_displayed=strings_displayed,
            strings_displayed_in_iframe=strings_displayed_in_iframe,
        )

    def test_0025_load_view_page_for_previous_revision(self):
        """Load a citable url for a past changeset revision and verify that strings display.

        We are at step 5.
        Visit the following url and check for appropriate strings: ``<tool shed base url>/view/user1/filtering_0420/<revision 0>`` .
        The resulting page should not contain change set revision 1, but should contain change set revision 0.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        encoded_user_id = self.security.encode_id(test_user_1.id)
        encoded_repository_id = repository.id
        # Since twill does not load the contents of an iframe, we need to check that the iframe has been generated correctly,
        # then directly load the url that the iframe should be loading and check for the expected strings.
        # The iframe should point to /repository/view_repository?id=<encoded repository ID>
        if self.is_v2:
            strings_displayed = []
        else:
            strings_displayed = ["/repository", "view_repository", f"id={encoded_repository_id}"]
        strings_displayed_in_iframe = [
            "user1",
            "filtering_0420",
            self._escape_page_content_if_needed(repository_long_description),
            first_changeset_hash,
        ]
        if not self.is_v2:
            strings_displayed_in_iframe.append("Link to this repository revision:")
            strings_displayed_in_iframe.append(f"{self.url}/view/user1/filtering_0420/{first_changeset_hash}")
        self.load_citable_url(
            username="user1",
            repository_name="filtering_0420",
            changeset_revision=first_changeset_hash,
            encoded_user_id=encoded_user_id,
            encoded_repository_id=encoded_repository_id,
            strings_displayed=strings_displayed,
            strings_displayed_in_iframe=strings_displayed_in_iframe,
        )

    def test_0030_load_sharable_url_with_invalid_changeset_revision(self):
        """Load a citable url with an invalid changeset revision specified."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        encoded_user_id = self.security.encode_id(test_user_1.id)
        encoded_repository_id = repository.id
        invalid_changeset_hash = "invalid"
        if not self.is_v2:
            # Since twill does not load the contents of an iframe, we need to check that the iframe has been generated correctly,
            # then directly load the url that the iframe should be loading and check for the expected strings.
            # The iframe should point to /repository/view_repository?id=<encoded repository ID>&status=error
            strings_displayed = ["/repository", "view_repository", f"id={encoded_repository_id}"]
            strings_displayed.extend(
                ["The+change+log", "does+not+include+revision", invalid_changeset_hash, "status=error"]
            )
        else:
            strings_displayed = ["The change log does not include revision " + invalid_changeset_hash]
        self.load_citable_url(
            username="user1",
            repository_name="filtering_0420",
            changeset_revision=invalid_changeset_hash,
            encoded_user_id=encoded_user_id,
            encoded_repository_id=encoded_repository_id,
            strings_displayed=strings_displayed,
        )

    def test_0035_load_sharable_url_with_invalid_repository_name(self):
        """Load a citable url with an invalid changeset revision specified.

        We are at step 7.
        Visit the following url and check for appropriate strings: <tool shed base url>/view/user1/!!invalid!!
        """
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        encoded_user_id = self.security.encode_id(test_user_1.id)
        # Since twill does not load the contents of an iframe, we need to check that the iframe has been generated correctly,
        # then directly load the url that the iframe should be loading and check for the expected strings.
        # The iframe should point to /repository/browse_repositories?user_id=<encoded user ID>&operation=repositories_by_user
        if not self.is_v2:
            strings_displayed = ["/repository", "browse_repositories", "user1"]
            strings_displayed.extend(
                ["list+of+repositories+owned", "does+not+include+one+named", "%21%21invalid%21%21", "status=error"]
            )
            strings_displayed_in_iframe = ["user1", "filtering_0420"]
            strings_displayed_in_iframe.append("Repositories Owned by user1")
        else:
            strings_displayed = ["Repository user1/!!invalid!! is not found"]
            strings_displayed_in_iframe = []
        self.load_citable_url(
            username="user1",
            repository_name="!!invalid!!",
            changeset_revision=None,
            encoded_user_id=encoded_user_id,
            encoded_repository_id=None,
            strings_displayed=strings_displayed,
            strings_displayed_in_iframe=strings_displayed_in_iframe,
        )

    def test_0040_load_sharable_url_with_invalid_owner(self):
        """Load a citable url with an invalid owner.

        We are at step 8.
        Visit the following url and check for appropriate strings: <tool shed base url>/view/!!invalid!!
        """
        if not self.is_v2:
            strings_displayed = ["The tool shed", self.url, "contains no repositories owned by", "!!invalid!!"]
        else:
            strings_displayed = ["No repositories found"]
        self.load_citable_url(
            username="!!invalid!!",
            repository_name=None,
            changeset_revision=None,
            encoded_user_id=None,
            encoded_repository_id=None,
            strings_displayed=strings_displayed,
        )



import logging
import os
import sys
from argparse import Namespace
from typing import Optional

from galaxy.model.migrations.base import (
    BaseCommand,
    BaseDbScript,
    BaseParserBuilder,
)
from tool_shed.webapp.model.migrations import verify_database
from tool_shed.webapp.model.migrations.scripts import (
    get_dburl,
    get_dburl_from_file,
)

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

CONFIG_FILE_ARG = "--toolshed-config"

# Update this dict with tags for each new release.
# Note: the key should NOT be a prefix of an existing revision hash in alembic/versions/.
# For example, if we have a hash 231xxxxxxxxx and use 231 as the key for release 23.1,
# then using 231 as a partial revision identifier like `sh manage_toolshed_db.sh upgrade 231`
# will map to release 23.1 instead of revision 231xxxxxxxxx.
REVISION_TAGS = {
    "release_23.1": "1b5bf427db25",
    "23.1": "1b5bf427db25",
    "release_23.2": "1b5bf427db25",
    "23.2": "1b5bf427db25",
    "release_24.0": "1b5bf427db25",
    "24.0": "1b5bf427db25",
    "release_24.1": "1b5bf427db25",
    "24.1": "1b5bf427db25",
}


class ParserBuilder(BaseParserBuilder):
    def _get_command_object(self):
        return Command()


class Command(BaseCommand):
    def init(self, args: Namespace) -> None:
        url = get_dburl(sys.argv, os.getcwd())
        verify_database(url)

    def _get_dbscript(self, config_file: str) -> BaseDbScript:
        return DbScript(config_file)


class DbScript(BaseDbScript):
    def _set_dburl(self, config_file: Optional[str] = None) -> None:
        self.url = get_dburl_from_file(os.getcwd(), config_file)
        self.alembic_config.set_main_option("sqlalchemy.url", self.url)

    def _upgrade_to_head(self, is_sql_mode: bool) -> None:
        self.alembic_config.set_main_option("sqlalchemy.url", self.url)
        self._upgrade_to_revision("head", is_sql_mode)

    def _revision_tags(self):
        return REVISION_TAGS



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_maker_repository_name = "column_maker_0110"
column_maker_repository_description = "A flexible aligner."
column_maker_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

emboss_repository_name = "emboss_0110"
emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools"

category_name = "Test 0110 Invalid Repository Dependencies"
category_desc = "Test 0110 Invalid Repository Dependencies"
running_standalone = False


class TestBasicRepositoryDependencies(ShedTwillTestCase):
    """Testing emboss 5 with repository dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_category(self):
        """Create a category for this test suite"""
        self.create_category(name=category_name, description=category_desc)

    def test_0010_create_emboss_dependendent_column_maker_repository_and_upload_tarball(self):
        """Create and populate the column_maker repository."""
        global running_standalone
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        category = self.populator.get_category_with_name(category_name)
        column_maker_repository = self.get_or_create_repository(
            name=column_maker_repository_name,
            description=column_maker_repository_description,
            long_description=column_maker_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(column_maker_repository):
            running_standalone = True
            self.commit_tar_to_repository(
                column_maker_repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0020_create_emboss_5_repository_and_upload_files(self):
        """Create and populate the emboss_5_0110 repository."""
        if running_standalone:
            category = self.populator.get_category_with_name(category_name)
            repository = self.get_or_create_repository(
                name=emboss_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss tool tarball.",
            )

    def test_0025_generate_repository_dependency_with_invalid_url(self):
        """Generate a repository dependency for emboss 5 with an invalid URL."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_1110", additional_paths=["simple"])
            column_maker_repository = self._get_repository_by_name_and_owner(
                column_maker_repository_name, common.test_user_1_name
            )
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            url = "http://http://this is not an url!"
            name = column_maker_repository.name
            owner = column_maker_repository.owner
            changeset_revision = self.get_repository_tip(column_maker_repository)
            strings_displayed = ["Repository dependencies are currently supported only within the same tool shed"]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=emboss_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=False,
            )

    def test_0030_generate_repository_dependency_with_invalid_name(self):
        """Generate a repository dependency for emboss 5 with an invalid name."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_1110", additional_paths=["simple"])
            repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            url = self.url
            name = "!?invalid?!"
            owner = repository.owner
            changeset_revision = self.get_repository_tip(repository)
            strings_displayed = ["because the name is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=emboss_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=False,
            )

    def test_0035_generate_repository_dependency_with_invalid_owner(self):
        """Generate a repository dependency for emboss 5 with an invalid owner."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_1110", additional_paths=["simple"])
            repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            url = self.url
            name = repository.name
            owner = "!?invalid?!"
            changeset_revision = self.get_repository_tip(repository)
            strings_displayed = ["because the owner is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=emboss_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=False,
            )

    def test_0040_generate_repository_dependency_with_invalid_changeset_revision(self):
        """Generate a repository dependency for emboss 5 with an invalid changeset revision."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_1110", additional_paths=["simple", "invalid"])
            repository = self._get_repository_by_name_and_owner(column_maker_repository_name, common.test_user_1_name)
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            url = self.url
            name = repository.name
            owner = repository.owner
            changeset_revision = "!?invalid?!"
            strings_displayed = ["because the changeset revision is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=emboss_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=False,
            )

    def test_0045_install_repository_with_invalid_repository_dependency(self):
        """Install the repository and verify that galaxy detects invalid repository dependencies."""
        repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        preview_strings_displayed = [
            "emboss_0110",
            self.get_repository_tip(repository),
            "Ignoring repository dependency definition",
        ]
        self._install_repository(
            emboss_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            preview_strings_displayed=preview_strings_displayed,
        )
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        json = self.display_installed_repository_manage_json(installed_repository)
        assert "repository_dependencies" not in json



"""
Manager and Serializer for TS repositories.
"""

import json
import logging
from collections import namedtuple
from time import strftime
from typing import (
    Any,
    Callable,
    cast,
    Optional,
    Union,
)

from pydantic import BaseModel
from sqlalchemy import (
    false,
    func,
    or_,
    select,
)
from sqlalchemy.orm import scoped_session

from galaxy import web
from galaxy.exceptions import (
    ConfigDoesNotAllowException,
    InsufficientPermissionsException,
    InternalServerError,
    MalformedContents,
    ObjectNotFound,
    RequestParameterInvalidException,
)
from galaxy.security.idencoding import IdEncodingHelper
from galaxy.tool_shed.util import dependency_display
from galaxy.util import listify
from galaxy.util.tool_shed.encoding_util import tool_shed_encode
from tool_shed.context import (
    ProvidesRepositoriesContext,
    ProvidesUserContext,
)
from tool_shed.metadata import repository_metadata_manager
from tool_shed.repository_types import util as rt_util
from tool_shed.structured_app import ToolShedApp
from tool_shed.util import hg_util
from tool_shed.util.metadata_util import (
    get_all_dependencies,
    get_current_repository_metadata_for_changeset_revision,
    get_metadata_revisions,
    get_next_downloadable_changeset_revision,
    get_repository_metadata_by_changeset_revision,
)
from tool_shed.util.readme_util import build_readme_files_dict
from tool_shed.util.repository_content_util import upload_tar
from tool_shed.util.repository_util import (
    create_repository as low_level_create_repository,
    get_repo_info_dict,
    get_repositories_by_category,
    get_repository_in_tool_shed,
    validate_repository_name,
)
from tool_shed.util.shed_util_common import (
    count_repositories_in_category,
    get_category,
)
from tool_shed.util.tool_util import generate_message_for_invalid_tools
from tool_shed.webapp.model import (
    Repository,
    RepositoryCategoryAssociation,
    RepositoryMetadata,
    User,
)
from tool_shed.webapp.model.db import get_repository_by_name_and_owner
from tool_shed.webapp.search.repo_search import RepoSearch
from tool_shed_client.schema import (
    CreateRepositoryRequest,
    DetailedRepository,
    ExtraRepoInfo,
    IndexSortByType,
    LegacyInstallInfoTuple,
    PaginatedRepositoryIndexResults,
    Repository as SchemaRepository,
    RepositoryMetadataInstallInfoDict,
    RepositoryRevisionMetadata,
    ResetMetadataOnRepositoriesRequest,
    ResetMetadataOnRepositoriesResponse,
    ResetMetadataOnRepositoryResponse,
)
from .categories import get_value_mapper as category_value_mapper

log = logging.getLogger(__name__)


def search(trans: ProvidesUserContext, q: str, page: int = 1, page_size: int = 10):
    """
    Perform the search over TS repositories.
    Note that search works over the Whoosh index which you have
    to pre-create with scripts/tool_shed/build_ts_whoosh_index.sh manually.
    Also TS config option toolshed_search_on has to be True and
    whoosh_index_dir has to be specified.
    """
    app = trans.app
    conf = app.config
    if not conf.toolshed_search_on:
        raise ConfigDoesNotAllowException("Searching the TS through the API is turned off for this instance.")
    if not conf.whoosh_index_dir:
        raise ConfigDoesNotAllowException(
            "There is no directory for the search index specified. Please contact the administrator."
        )

    search_term = q.strip()
    if len(search_term) < 1:
        raise RequestParameterInvalidException("The search term has to be at least one character long.")

    repo_search = RepoSearch()

    Boosts = namedtuple(
        "Boosts",
        [
            "repo_name_boost",
            "repo_description_boost",
            "repo_long_description_boost",
            "repo_homepage_url_boost",
            "repo_remote_repository_url_boost",
            "categories_boost",
            "repo_owner_username_boost",
        ],
    )
    boosts = Boosts(
        float(conf.get("repo_name_boost", 0.9)),
        float(conf.get("repo_description_boost", 0.6)),
        float(conf.get("repo_long_description_boost", 0.5)),
        float(conf.get("repo_homepage_url_boost", 0.3)),
        float(conf.get("repo_remote_repository_url_boost", 0.2)),
        float(conf.get("categories_boost", 0.5)),
        float(conf.get("repo_owner_username_boost", 0.3)),
    )

    results = repo_search.search(trans, search_term, page, page_size, boosts)
    results["hostname"] = deprecated_hostname()
    return results


def deprecated_hostname() -> str:
    return web.url_for("/", qualified=True)


class UpdatesRequest(BaseModel):
    name: Optional[str] = None
    owner: Optional[str] = None
    changeset_revision: str
    hexlify: bool = True


def check_updates(app: ToolShedApp, request: UpdatesRequest) -> Union[str, dict[str, Any]]:
    name = request.name
    owner = request.owner
    changeset_revision = request.changeset_revision
    hexlify_this = request.hexlify
    repository = get_repository_by_name_and_owner(
        app.model.context, name, owner, eagerload_columns=[Repository.downloadable_revisions]
    )
    if repository and repository.downloadable_revisions:
        repository_metadata = get_repository_metadata_by_changeset_revision(
            app, app.security.encode_id(repository.id), changeset_revision
        )
        tool_shed_status_dict = {}
        # Handle repository deprecation.
        tool_shed_status_dict["repository_deprecated"] = str(repository.deprecated)
        tip_revision = repository.downloadable_revisions[0]
        # Handle latest installable revision.
        if changeset_revision == tip_revision:
            tool_shed_status_dict["latest_installable_revision"] = "True"
        else:
            next_installable_revision = get_next_downloadable_changeset_revision(app, repository, changeset_revision)
            if repository_metadata is None:
                if next_installable_revision and next_installable_revision != changeset_revision:
                    tool_shed_status_dict["latest_installable_revision"] = "True"
                else:
                    tool_shed_status_dict["latest_installable_revision"] = "False"
            else:
                if next_installable_revision and next_installable_revision != changeset_revision:
                    tool_shed_status_dict["latest_installable_revision"] = "False"
                else:
                    tool_shed_status_dict["latest_installable_revision"] = "True"
        # Handle revision updates.
        if changeset_revision == tip_revision:
            tool_shed_status_dict["revision_update"] = "False"
        else:
            if repository_metadata is None:
                tool_shed_status_dict["revision_update"] = "True"
            else:
                tool_shed_status_dict["revision_update"] = "False"
        # Handle revision upgrades.
        metadata_revisions = [revision[1] for revision in get_metadata_revisions(app, repository)]
        num_metadata_revisions = len(metadata_revisions)
        for index, metadata_revision in enumerate(metadata_revisions):
            if index == num_metadata_revisions:
                tool_shed_status_dict["revision_upgrade"] = "False"
                break
            if metadata_revision == changeset_revision:
                if num_metadata_revisions - index > 1:
                    tool_shed_status_dict["revision_upgrade"] = "True"
                else:
                    tool_shed_status_dict["revision_upgrade"] = "False"
                break
        return tool_shed_encode(tool_shed_status_dict) if hexlify_this else json.dumps(tool_shed_status_dict)
    return tool_shed_encode({}) if hexlify_this else json.dumps({})


def guid_to_repository(app: ToolShedApp, tool_id: str) -> Repository:
    # tool_id = remove_protocol_and_user_from_clone_url(tool_id)
    shed, _, owner, name, rest = tool_id.split("/", 5)
    return _get_repository_by_name_and_owner(app.model.context, name, owner)


def index_tool_ids(app: ToolShedApp, tool_ids: list[str]) -> dict[str, Any]:
    repository_found = []
    all_metadata = {}
    for tool_id in tool_ids:
        repository = guid_to_repository(app, tool_id)
        owner = repository.user.username
        name = repository.name
        assert name
        repository = _get_repository_by_name_and_owner(app.model.session, name, owner)
        if not repository:
            log.warning(f"Repository {owner}/{name} does not exist, skipping")
            continue
        for changeset, changehash in repository.installable_revisions(app):
            metadata = get_current_repository_metadata_for_changeset_revision(app, repository, changehash)
            tools: Optional[list[dict[str, Any]]] = metadata.metadata.get("tools")
            if not tools:
                log.warning(f"Repository {owner}/{name}/{changehash} does not contain valid tools, skipping")
                continue
            for tool_metadata in tools:
                if tool_metadata["guid"] in tool_ids:
                    repository_found.append(f"{int(changeset)}:{changehash}")
            metadata = get_current_repository_metadata_for_changeset_revision(app, repository, changehash)
            if metadata is None:
                continue
            metadata_dict = metadata.to_dict(
                value_mapper={"id": app.security.encode_id, "repository_id": app.security.encode_id}
            )
            metadata_dict["repository"] = repository.to_dict(value_mapper={"id": app.security.encode_id})
            if metadata.has_repository_dependencies:
                metadata_dict["repository_dependencies"] = get_all_dependencies(
                    app, metadata, processed_dependency_links=[]
                )
            else:
                metadata_dict["repository_dependencies"] = []
            if metadata.includes_tool_dependencies:
                metadata_dict["tool_dependencies"] = repository.get_tool_dependencies(app, changehash)
            else:
                metadata_dict["tool_dependencies"] = {}
            if metadata.includes_tools:
                metadata_dict["tools"] = metadata.metadata["tools"]
            all_metadata[f"{int(changeset)}:{changehash}"] = metadata_dict
    if repository_found:
        all_metadata["current_changeset"] = repository_found[0]
        # all_metadata[ 'found_changesets' ] = repository_found
        return all_metadata
    else:
        return {}


class IndexRequest(BaseModel):
    name: Optional[str] = None
    owner: Optional[str] = None
    deleted: bool = False
    filter: Optional[str] = None
    category_id: Optional[str] = None
    sort_by: IndexSortByType = "name"
    sort_desc: bool = False


class PaginatedIndexRequest(IndexRequest):
    page: int
    page_size: int


def index_repositories(app: ToolShedApp, index_request: IndexRequest) -> list[Repository]:
    session = app.model.context
    return list(session.scalars(_get_repositories_by_name_and_owner_and_deleted(app.security, index_request)))


def index_repositories_paginated(
    app: ToolShedApp, index_request: PaginatedIndexRequest
) -> PaginatedRepositoryIndexResults:
    session = app.model.context
    stmt = _get_repositories_by_name_and_owner_and_deleted(app.security, index_request)
    total_results = session.scalar(select(func.count()).select_from(stmt.subquery()))
    stmt = stmt.limit(index_request.page_size).offset((index_request.page - 1) * index_request.page_size)
    results = (to_model(app, r) for r in session.scalars(stmt).all())
    return PaginatedRepositoryIndexResults(
        total_results=total_results,
        page=index_request.page,
        page_size=index_request.page_size,
        hits=list(results),
        hostname=deprecated_hostname(),
    )


def can_manage_repo(trans: ProvidesUserContext, repository: Repository) -> bool:
    security_agent = trans.app.security_agent
    return trans.user_is_admin or security_agent.user_can_administer_repository(trans.user, repository)


def can_update_repo(trans: ProvidesUserContext, repository: Repository) -> bool:
    app = trans.app
    security_agent = app.security_agent
    return can_manage_repo(trans, repository) or security_agent.can_push(app, trans.user, repository)


def get_repository_metadata_for_management(
    trans: ProvidesUserContext, encoded_repository_id: str, changeset_revision: str
) -> RepositoryMetadata:
    repository = get_repository_in_tool_shed(trans.app, encoded_repository_id)
    ensure_can_manage(trans, repository, "Cannot manage target repository")
    revisions = [r for r in repository.metadata_revisions if r.changeset_revision == changeset_revision]
    if len(revisions) != 1:
        raise ObjectNotFound()
    repository_metadata = revisions[0]
    return repository_metadata


def get_install_info(trans: ProvidesRepositoriesContext, name, owner, changeset_revision) -> LegacyInstallInfoTuple:
    app = trans.app
    value_mapper = get_value_mapper(app)
    # Example URL:
    # http://<xyz>/api/repositories/get_repository_revision_install_info?name=<n>&owner=<o>&changeset_revision=<cr>
    if name and owner and changeset_revision:
        # Get the repository information.
        repository = get_repository_by_name_and_owner(
            app.model.context, name, owner, eagerload_columns=[Repository.downloadable_revisions]
        )
        if repository is None:
            log.debug(f"Cannot locate repository {name} owned by {owner}")
            return {}, {}, {}
        encoded_repository_id = app.security.encode_id(repository.id)
        repository_dict: dict = repository.to_dict(view="element", value_mapper=value_mapper)
        repository_dict["url"] = web.url_for(controller="repositories", action="show", id=encoded_repository_id)
        # Get the repository_metadata information.
        repository_metadata = get_repository_metadata_by_changeset_revision(
            app, encoded_repository_id, changeset_revision
        )
        if repository_metadata is None:
            # The changeset_revision column in the repository_metadata table has been updated with a new
            # value value, so find the changeset_revision to which we need to update.
            new_changeset_revision = get_next_downloadable_changeset_revision(app, repository, changeset_revision)
            repository_metadata = get_repository_metadata_by_changeset_revision(
                app, encoded_repository_id, new_changeset_revision
            )
            changeset_revision = new_changeset_revision
        if repository_metadata is not None:
            encoded_repository_metadata_id = app.security.encode_id(repository_metadata.id)
            repository_metadata_dict: RepositoryMetadataInstallInfoDict = cast(
                RepositoryMetadataInstallInfoDict,
                repository_metadata.to_dict(view="collection", value_mapper=value_mapper),
            )
            repository_metadata_dict["url"] = web.url_for(
                controller="repository_revisions", action="show", id=encoded_repository_metadata_id
            )
            if "tools" in repository_metadata.metadata:
                repository_metadata_dict["valid_tools"] = repository_metadata.metadata["tools"]
            # Get the repo_info_dict for installing the repository.
            repo_info_dict: ExtraRepoInfo
            (
                repo_info_dict,
                includes_tools,
                includes_tool_dependencies,
                includes_tools_for_display_in_tool_panel,
                has_repository_dependencies,
                has_repository_dependencies_only_if_compiling_contained_td,
            ) = get_repo_info_dict(trans, encoded_repository_id, changeset_revision)
            return repository_dict, repository_metadata_dict, repo_info_dict
        else:
            log.debug(
                "Unable to locate repository_metadata record for repository id %s and changeset_revision %s",
                repository.id,
                changeset_revision,
            )
            return repository_dict, {}, {}
    else:
        debug_msg = "Error in the Tool Shed repositories API in get_repository_revision_install_info: "
        debug_msg += f"Invalid name {name} or owner {owner} or changeset_revision {changeset_revision} received."
        log.debug(debug_msg)
        return {}, {}, {}


def get_value_mapper(app: ToolShedApp) -> dict[str, Callable]:
    value_mapper = {
        "id": app.security.encode_id,
        "repository_id": app.security.encode_id,
        "user_id": app.security.encode_id,
    }
    return value_mapper


def get_ordered_installable_revisions(
    app: ToolShedApp, name: Optional[str], owner: Optional[str], tsr_id: Optional[str]
) -> list[str]:
    eagerload_columns = [Repository.downloadable_revisions]
    if None not in [name, owner]:
        # Get the repository information.
        repository = get_repository_by_name_and_owner(
            app.model.context, name, owner, eagerload_columns=eagerload_columns
        )
        if repository is None:
            raise ObjectNotFound(f"No repository named {name} found with owner {owner}")
    elif tsr_id is not None:
        repository = get_repository_in_tool_shed(app, tsr_id, eagerload_columns=eagerload_columns)
    else:
        error_message = "Error in the Tool Shed repositories API in get_ordered_installable_revisions: "
        error_message += "invalid parameters received."
        log.debug(error_message)
        return []
    return [revision[1] for revision in repository.installable_revisions(app, sort_revisions=True)]


def get_repository_metadata_dict(app: ToolShedApp, id: str, recursive: bool, downloadable_only: bool) -> dict[str, Any]:
    all_metadata = {}
    repository = get_repository_in_tool_shed(app, id, eagerload_columns=[Repository.downloadable_revisions])
    for changeset, changehash in get_metadata_revisions(
        app, repository, sort_revisions=True, downloadable=downloadable_only
    ):
        metadata = get_current_repository_metadata_for_changeset_revision(app, repository, changehash)
        if metadata is None:
            continue
        metadata_dict = get_repository_revision_metadata_dict(app, repository, metadata, recursive=recursive)
        all_metadata[f"{int(changeset)}:{changehash}"] = metadata_dict
    return all_metadata


def get_repository_revision_metadata_dict(
    app: ToolShedApp, repository: Repository, metadata: RepositoryMetadata, recursive: bool = False
):
    metadata_dict = metadata.to_dict(
        value_mapper={"id": app.security.encode_id, "repository_id": app.security.encode_id}
    )
    metadata_dict["repository"] = repository.to_dict(
        value_mapper={"id": app.security.encode_id, "user_id": app.security.encode_id}
    )
    if metadata.has_repository_dependencies and recursive:
        metadata_dict["repository_dependencies"] = get_all_dependencies(app, metadata, processed_dependency_links=[])
    else:
        metadata_dict["repository_dependencies"] = []
    if metadata.includes_tools:
        metadata_dict["tools"] = metadata.metadata["tools"]
    metadata_dict["invalid_tools"] = metadata.metadata.get("invalid_tools", [])
    return metadata_dict


def get_repository_revision_metadata_model(
    app: ToolShedApp, repository: Repository, metadata: RepositoryMetadata, recursive: bool = False
) -> RepositoryRevisionMetadata:
    as_dict = get_repository_revision_metadata_dict(app, repository, metadata, recursive=recursive)
    return RepositoryRevisionMetadata(**as_dict)


def readmes(app: ToolShedApp, repository: Repository, changeset_revision: str) -> dict:
    encoded_repository_id = app.security.encode_id(repository.id)
    repository_metadata = get_repository_metadata_by_changeset_revision(app, encoded_repository_id, changeset_revision)
    if repository_metadata:
        metadata = repository_metadata.metadata
        if metadata:
            return build_readme_files_dict(app, repository, changeset_revision, repository_metadata.metadata)
    return {}


def reset_metadata_on_repository(trans: ProvidesUserContext, repository_id) -> ResetMetadataOnRepositoryResponse:
    app: ToolShedApp = trans.app

    def handle_repository(trans, start_time, repository):
        results = dict(start_time=start_time, repository_status=[])
        try:
            rmm = repository_metadata_manager.RepositoryMetadataManager(
                trans,
                repository=repository,
                resetting_all_metadata_on_repository=True,
                updating_installed_repository=False,
                persist=False,
            )
            rmm.reset_all_metadata_on_repository_in_tool_shed()
            rmm_invalid_file_tups = rmm.get_invalid_file_tups()
            if rmm_invalid_file_tups:
                message = generate_message_for_invalid_tools(
                    app, rmm_invalid_file_tups, repository, None, as_html=False
                )
                results["status"] = "warning"
            else:
                message = (
                    f"Successfully reset metadata on repository {repository.name} owned by {repository.user.username}"
                )
                results["status"] = "ok"
        except Exception as e:
            message = (
                f"Error resetting metadata on repository {repository.name} owned by {repository.user.username}: {e}"
            )
            results["status"] = "error"
        status = f"{repository.name} : {message}"
        results["repository_status"].append(status)
        return results

    if repository_id is not None:
        repository = get_repository_in_tool_shed(app, repository_id)
        start_time = strftime("%Y-%m-%d %H:%M:%S")
        log.debug(f"{start_time}...resetting metadata on repository {repository.name}")
        results = handle_repository(trans, start_time, repository)
        stop_time = strftime("%Y-%m-%d %H:%M:%S")
        results["stop_time"] = stop_time
    return ResetMetadataOnRepositoryResponse(**results)


def reset_metadata_on_repositories(
    trans: ProvidesRepositoriesContext, request: ResetMetadataOnRepositoriesRequest
) -> ResetMetadataOnRepositoriesResponse:

    def handle_repository(trans, repository, results):
        log.debug(f"Resetting metadata on repository {repository.name}")
        try:
            rmm = repository_metadata_manager.RepositoryMetadataManager(
                trans,
                resetting_all_metadata_on_repository=True,
                updating_installed_repository=False,
                repository=repository,
                persist=False,
            )
            rmm.reset_all_metadata_on_repository_in_tool_shed()
            rmm_invalid_file_tups = rmm.get_invalid_file_tups()
            if rmm_invalid_file_tups:
                message = generate_message_for_invalid_tools(
                    trans.app, rmm_invalid_file_tups, repository, None, as_html=False
                )
                results["unsuccessful_count"] += 1
            else:
                message = (
                    f"Successfully reset metadata on repository {repository.name} owned by {repository.user.username}"
                )
                results["successful_count"] += 1
        except Exception as e:
            message = (
                f"Error resetting metadata on repository {repository.name} owned by {repository.user.username}: {e}"
            )
            results["unsuccessful_count"] += 1
        status = f"{repository.name} : {message}"
        results["repository_status"].append(status)
        return results

    start_time = strftime("%Y-%m-%d %H:%M:%S")
    results = dict(start_time=start_time, repository_status=[], successful_count=0, unsuccessful_count=0)
    handled_repository_ids: list[str] = []
    encoded_ids_to_skip = request.encoded_ids_to_skip or []
    if trans.user_is_admin:
        my_writable = request.my_writable
    else:
        my_writable = True
    rmm = repository_metadata_manager.RepositoryMetadataManager(
        trans,
        resetting_all_metadata_on_repository=True,
        updating_installed_repository=False,
        persist=False,
    )
    # First reset metadata on all repositories of type repository_dependency_definition.
    for repository in rmm.get_repositories_for_setting_metadata(my_writable=my_writable, order=False):
        encoded_id = trans.security.encode_id(repository.id)
        if encoded_id in encoded_ids_to_skip:
            log.debug(
                "Skipping repository with id %s because it is in encoded_ids_to_skip %s",
                repository.id,
                encoded_ids_to_skip,
            )
        elif repository.type == rt_util.TOOL_DEPENDENCY_DEFINITION and repository.id not in handled_repository_ids:
            results = handle_repository(trans, repository, results)
    # Now reset metadata on all remaining repositories.
    for repository in rmm.get_repositories_for_setting_metadata(my_writable=my_writable, order=False):
        encoded_id = trans.security.encode_id(repository.id)
        if encoded_id in encoded_ids_to_skip:
            log.debug(
                "Skipping repository with id %s because it is in encoded_ids_to_skip %s",
                repository.id,
                encoded_ids_to_skip,
            )
        elif repository.type != rt_util.TOOL_DEPENDENCY_DEFINITION and repository.id not in handled_repository_ids:
            results = handle_repository(trans, repository, results)
    stop_time = strftime("%Y-%m-%d %H:%M:%S")
    results["stop_time"] = stop_time
    return ResetMetadataOnRepositoriesResponse(**results)


def create_repository(trans: ProvidesUserContext, request: CreateRepositoryRequest) -> Repository:
    app: ToolShedApp = trans.app
    user = trans.user
    assert user
    category_ids = listify(request.category_ids)
    name = request.name
    if invalid_message := validate_repository_name(app, name, user):
        raise RequestParameterInvalidException(invalid_message)

    repo, _ = low_level_create_repository(
        app=app,
        name=name,
        type=request.type_,
        description=request.synopsis,
        long_description=request.description,
        user=user,
        category_ids=category_ids,
        remote_repository_url=request.remote_repository_url,
        homepage_url=request.homepage_url,
    )
    return repo


def to_element_dict(app, repository: Repository, include_categories: bool = False) -> dict[str, Any]:
    value_mapper = get_value_mapper(app)
    repository_dict = repository.to_dict(view="element", value_mapper=value_mapper)
    if include_categories:
        repository_dict["category_ids"] = [app.security.encode_id(x.category.id) for x in repository.categories]
    return repository_dict


def repositories_by_category(
    app: ToolShedApp,
    category_id: str,
    page: Optional[int] = None,
    sort_key: str = "name",
    sort_order: str = "asc",
    installable: bool = True,
):
    category = get_category(app, category_id)
    category_dict: dict[str, Any]
    if category is None:
        category_dict = dict(message=f"Unable to locate category record for id {str(id)}.", status="error")
        return category_dict
    category_dict = category.to_dict(view="element", value_mapper=category_value_mapper(app))
    category_dict["repository_count"] = count_repositories_in_category(app, category_id)
    repositories = get_repositories_by_category(
        app, category.id, installable=installable, sort_order=sort_order, sort_key=sort_key, page=page
    )
    category_dict["repositories"] = repositories
    return category_dict


def to_model(app, repository: Repository) -> SchemaRepository:
    return SchemaRepository(**to_element_dict(app, repository))


def to_detailed_model(app, repository: Repository) -> DetailedRepository:
    return DetailedRepository(**to_element_dict(app, repository))


def upload_tar_and_set_metadata(
    trans: ProvidesRepositoriesContext,
    host: str,
    repository: Repository,
    uploaded_file,
    commit_message: str,
    dry_run: bool = False,
):
    app = trans.app
    user = trans.user
    assert user
    assert user.username
    repo_dir = repository.repo_path(app)
    tip = repository.tip()
    tar_response = upload_tar(
        trans,
        user.username,
        repository,
        uploaded_file,
        commit_message,
    )
    (
        ok,
        message,
        _,
        content_alert_str,
        _,
        _,
    ) = tar_response
    if ok:
        # Update the repository files for browsing.
        hg_util.update_repository(repo_dir)
        # Get the new repository tip.
        if tip == repository.tip():
            raise MalformedContents("No changes to repository.")
        else:
            rmm = repository_metadata_manager.RepositoryMetadataManager(trans, repository=repository)
            _, error_message = rmm.set_repository_metadata_due_to_new_tip(host, content_alert_str=content_alert_str)
            if error_message:
                raise InternalServerError(error_message)
            dd = dependency_display.DependencyDisplayer(app)
            if str(repository.type) not in [
                rt_util.REPOSITORY_SUITE_DEFINITION,
                rt_util.TOOL_DEPENDENCY_DEFINITION,
            ]:
                # Provide a warning message if a tool_dependencies.xml file is provided, but tool dependencies
                # weren't loaded due to a requirement tag mismatch or some other problem.  Tool dependency
                # definitions can define orphan tool dependencies (no relationship to any tools contained in the
                # repository), so warning messages are important because orphans are always valid.  The repository
                # owner must be warned in case they did not intend to define an orphan dependency, but simply
                # provided incorrect information (tool shed, name owner, changeset_revision) for the definition.
                if repository.metadata_revisions:
                    # A repository's metadata revisions are order descending by update_time, so the zeroth revision
                    # will be the tip just after an upload.
                    metadata_dict = repository.metadata_revisions[0].metadata
                else:
                    metadata_dict = {}
                orphan_message = dd.generate_message_for_orphan_tool_dependencies(repository, metadata_dict)
                if orphan_message:
                    message += orphan_message
    else:
        raise InternalServerError(message)
    return message


def ensure_can_manage(trans: ProvidesUserContext, repository: Repository, error_message: Optional[str] = None) -> None:
    if not can_manage_repo(trans, repository):
        error_message = error_message or "You do not have permission to update this repository."
        raise InsufficientPermissionsException(error_message)


def _get_repository_by_name_and_owner(session: scoped_session, name: str, owner: str):
    stmt = (
        select(Repository)
        .where(Repository.deprecated == false())
        .where(Repository.deleted == false())
        .where(Repository.name == name)
        .where(User.username == owner)
        .where(Repository.user_id == User.id)
        .limit(1)
    )
    return session.scalars(stmt).first()


def _get_repositories_by_name_and_owner_and_deleted(security: IdEncodingHelper, index_request: IndexRequest):
    owner = index_request.owner
    name = index_request.name
    deleted = index_request.deleted
    filter = index_request.filter
    stmt = select(Repository).where(Repository.deprecated == false()).where(Repository.deleted == deleted)
    if owner is not None or filter:
        stmt = stmt.join(Repository.user)
    if owner is not None:
        stmt = stmt.where(User.username == owner)
    if name is not None:
        stmt = stmt.where(Repository.name == name)
    if filter:
        filter_ilike_str = f"%{filter}%"
        stmt = stmt.where(
            or_(
                User.username.ilike(filter_ilike_str),
                Repository.name.ilike(filter_ilike_str),
                Repository.description.ilike(filter_ilike_str),
            )
        )
    if index_request.category_id is not None:
        category_id = security.decode_id(index_request.category_id)
        stmt = stmt.where(RepositoryCategoryAssociation.category_id == category_id)
        stmt = stmt.where(RepositoryCategoryAssociation.repository_id == Repository.id)
    sort_by_str = index_request.sort_by
    sort_desc = index_request.sort_desc
    sort_by: Any
    if sort_by_str == "name":
        sort_by = Repository.name
    elif sort_by_str == "create_time":
        sort_by = Repository.create_time
    if sort_desc:
        sort_by = sort_by.desc()
    stmt = stmt.order_by(sort_by)
    return stmt



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

repository_name = "freebayes_0010"
repository_description = "Galaxy's freebayes tool"
repository_long_description = "Long description of Galaxy's freebayes tool"
category_name = "Test 0010 Repository With Tool Dependencies"
log = logging.getLogger(__name__)


class TestToolWithToolDependencies(ShedTwillTestCase):
    """Test installing a repository with tool dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_repositories_and_categories_exist(self):
        """Create the 0010 category and upload the freebayes repository to it, if necessary."""
        category = self.create_category(
            name=category_name, description="Tests for a repository with tool dependencies."
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
        )
        if self.repository_is_new(repository):
            self.setup_freebayes_0010_repo(repository)

    def test_0010_browse_tool_shed(self):
        """Browse the available tool sheds in this Galaxy instance and preview the freebayes tool."""
        self.browse_tool_shed(url=self.url, strings_displayed=[category_name])
        category = self.populator.get_category_with_name(category_name)
        self.browse_category(category, strings_displayed=[repository_name])
        if not self.is_v2:
            strings_displayed = [repository_name, "Valid tools", "Tool dependencies"]
            self.preview_repository_in_tool_shed(
                repository_name, common.test_user_1_name, strings_displayed=strings_displayed
            )

    def test_0015_install_freebayes_repository(self):
        """Install the freebayes repository without installing tool dependencies."""
        self._install_repository(
            repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            new_tool_panel_section_label="test_1010",
        )
        installed_repository = self._get_installed_repository_by_name_owner(repository_name, common.test_user_1_name)
        assert self._get_installed_repository_for(
            common.test_user_1, repository_name, installed_repository.installed_changeset_revision
        )
        self._assert_has_valid_tool_with_name("FreeBayes")
        self._assert_repo_has_tool_with_id(installed_repository, "freebayes")

    def test_0020_verify_installed_repository_metadata(self):
        """Verify that resetting the metadata on an installed repository does not change the metadata."""
        self.verify_installed_repository_metadata_unchanged(repository_name, common.test_user_1_name)

    def test_0025_verify_sample_files(self):
        """Verify that the installed repository populated shed_tool_data_table.xml and the sample files."""
        self.verify_installed_repository_data_table_entries(required_data_table_entries=["sam_fa_indexes"])
